# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper
from newrelic.core.otlp_utils import otlp_content_setting

if otlp_content_setting == "protobuf":
    from google.protobuf.json_format import MessageToDict
else:
    MessageToDict = None


def data_points_to_dict(data_points):
    return {
        frozenset(
            {attr["key"]: attribute_to_value(attr["value"]) for attr in (data_point.get("attributes") or [])}.items()
        )
        or None: data_point
        for data_point in data_points
    }


def attribute_to_value(attribute):
    attribute_type, attribute_value = next(iter(attribute.items()))
    if attribute_type == "int_value":
        return int(attribute_value)
    elif attribute_type == "double_value":
        return float(attribute_value)
    elif attribute_type == "bool_value":
        return bool(attribute_value)
    elif attribute_type == "string_value":
        return str(attribute_value)
    else:
        raise TypeError("Invalid attribute type: %s" % attribute_type)


def payload_to_metrics(payload):
    if type(payload) is not dict:
        message = MessageToDict(payload, use_integers_for_enums=True, preserving_proto_field_name=True)
    else:
        message = payload

    resource_metrics = message.get("resource_metrics")
    assert len(resource_metrics) == 1
    resource_metrics = resource_metrics[0]

    resource = resource_metrics.get("resource")
    assert resource and resource.get("attributes")[0] == {
        "key": "instrumentation.provider",
        "value": {"string_value": "newrelic-opentelemetry-python-ml"},
    }
    scope_metrics = resource_metrics.get("scope_metrics")
    assert len(scope_metrics) == 1
    scope_metrics = scope_metrics[0]

    scope = scope_metrics.get("scope")
    assert scope is None
    metrics = scope_metrics.get("metrics")

    sent_summary_metrics = {}
    sent_count_metrics = {}
    for metric in metrics:
        metric_name = metric["name"]
        if metric.get("sum"):
            sent_count_metrics[metric_name] = metric
        elif metric.get("summary"):
            sent_summary_metrics[metric_name] = metric
        else:
            raise TypeError("Unknown metrics type for metric: %s" % metric)

    return sent_summary_metrics, sent_count_metrics


def validate_dimensional_metric_payload(summary_metrics=None, count_metrics=None):
    # Validates OTLP metrics as they are sent to the collector.

    summary_metrics = summary_metrics or []
    count_metrics = count_metrics or []

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        recorded_metrics = []

        @transient_function_wrapper("newrelic.core.agent_protocol", "OtlpProtocol.send")
        def send_request_wrapper(wrapped, instance, args, kwargs):
            def _bind_params(method, payload=(), *args, **kwargs):
                return method, payload

            method, payload = _bind_params(*args, **kwargs)

            if method == "dimensional_metric_data" and payload:
                recorded_metrics.append(payload)

            return wrapped(*args, **kwargs)

        wrapped = send_request_wrapper(wrapped)
        val = wrapped(*args, **kwargs)
        assert recorded_metrics

        decoded_payloads = [payload_to_metrics(payload) for payload in recorded_metrics]
        for sent_summary_metrics, sent_count_metrics in decoded_payloads:
            for metric, tags, count in summary_metrics:
                if isinstance(tags, dict):
                    tags = frozenset(tags.items())

                if not count:
                    if metric in sent_summary_metrics:
                        data_points = data_points_to_dict(sent_summary_metrics[metric]["summary"]["data_points"])
                        assert tags not in data_points, "(%s, %s) Unexpected but found." % (metric, tags and dict(tags))
                else:
                    assert metric in sent_summary_metrics, "%s Not Found. Got: %s" % (
                        metric,
                        list(sent_summary_metrics.keys()),
                    )
                    data_points = data_points_to_dict(sent_summary_metrics[metric]["summary"]["data_points"])
                    assert tags in data_points, "(%s, %s) Not Found. Got: %s" % (
                        metric,
                        tags and dict(tags),
                        list(data_points.keys()),
                    )

                    # Validate metric format
                    metric_container = data_points[tags]
                    for key in ("start_time_unix_nano", "time_unix_nano", "count", "sum", "quantile_values"):
                        assert key in metric_container, "Invalid metric format. Missing key: %s" % key
                    quantile_values = metric_container["quantile_values"]
                    assert len(quantile_values) == 2  # Min and Max

                    # Validate metric count
                    if count != "present":
                        assert int(metric_container["count"]) == count, "(%s, %s): Expected: %s Got: %s" % (
                            metric,
                            tags and dict(tags),
                            count,
                            metric_container["count"],
                        )

            for metric, tags, count in count_metrics:
                if isinstance(tags, dict):
                    tags = frozenset(tags.items())

                if not count:
                    if metric in sent_count_metrics:
                        data_points = data_points_to_dict(sent_count_metrics[metric]["sum"]["data_points"])
                        assert tags not in data_points, "(%s, %s) Unexpected but found." % (metric, tags and dict(tags))
                else:
                    assert metric in sent_count_metrics, "%s Not Found. Got: %s" % (
                        metric,
                        list(sent_count_metrics.keys()),
                    )
                    data_points = data_points_to_dict(sent_count_metrics[metric]["sum"]["data_points"])
                    assert tags in data_points, "(%s, %s) Not Found. Got: %s" % (
                        metric,
                        tags and dict(tags),
                        list(data_points.keys()),
                    )

                    # Validate metric format
                    assert sent_count_metrics[metric]["sum"].get("is_monotonic")
                    assert sent_count_metrics[metric]["sum"].get("aggregation_temporality") == 1
                    metric_container = data_points[tags]
                    for key in ("start_time_unix_nano", "time_unix_nano", "as_int"):
                        assert key in metric_container, "Invalid metric format. Missing key: %s" % key

                    # Validate metric count
                    if count != "present":
                        assert int(metric_container["as_int"]) == count, "(%s, %s): Expected: %s Got: %s" % (
                            metric,
                            tags and dict(tags),
                            count,
                            metric_container["count"],
                        )

        return val

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.datastore_trace import wrap_datastore_trace

_memcache_client_methods = (
    "set",
    "set_many",
    "add",
    "replace",
    "append",
    "prepend",
    "cas",
    "get",
    "get_many",
    "gets",
    "gets_many",
    "delete",
    "delete_many",
    "incr",
    "decr",
    "touch",
    "stats",
    "flush_all",
    "quit",
)


def instrument_pymemcache_client(module):
    for name in _memcache_client_methods:
        if hasattr(module.Client, name):
            wrap_datastore_trace(module, "Client.%s" % name, product="Memcached", target=None, operation=name)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper
from newrelic.core.otlp_utils import otlp_content_setting

if otlp_content_setting == "protobuf":
    from google.protobuf.json_format import MessageToDict
else:
    MessageToDict = None


def attribute_to_value(attribute):
    attribute_type, attribute_value = next(iter(attribute.items()))
    if attribute_type == "int_value":
        return int(attribute_value)
    elif attribute_type == "double_value":
        return float(attribute_value)
    elif attribute_type == "bool_value":
        return bool(attribute_value)
    elif attribute_type == "string_value":
        return str(attribute_value)
    else:
        raise TypeError("Invalid attribute type: %s" % attribute_type)


def payload_to_ml_events(payload):
    if type(payload) is not dict:
        message = MessageToDict(payload, use_integers_for_enums=True, preserving_proto_field_name=True)
    else:
        message = payload

    resource_logs = message.get("resource_logs")
    assert len(resource_logs) == 1
    resource_logs = resource_logs[0]
    resource = resource_logs.get("resource")
    assert resource and resource.get("attributes")[0] == {
        "key": "instrumentation.provider",
        "value": {"string_value": "newrelic-opentelemetry-python-ml"},
    }
    scope_logs = resource_logs.get("scope_logs")
    assert len(scope_logs) == 1
    scope_logs = scope_logs[0]

    scope = scope_logs.get("scope")
    assert scope is None
    logs = scope_logs.get("log_records")

    return logs


def validate_ml_event_payload(ml_events=None):
    # Validates OTLP events as they are sent to the collector.

    ml_events = ml_events or []

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        recorded_ml_events = []

        @transient_function_wrapper("newrelic.core.agent_protocol", "OtlpProtocol.send")
        def send_request_wrapper(wrapped, instance, args, kwargs):
            def _bind_params(method, payload=(), *args, **kwargs):
                return method, payload

            method, payload = _bind_params(*args, **kwargs)

            if method == "ml_event_data" and payload:
                recorded_ml_events.append(payload)

            return wrapped(*args, **kwargs)

        wrapped = send_request_wrapper(wrapped)
        val = wrapped(*args, **kwargs)
        assert recorded_ml_events

        decoded_payloads = [payload_to_ml_events(payload) for payload in recorded_ml_events]
        all_logs = []
        for sent_logs in decoded_payloads:
            for data_point in sent_logs:
                for key in ("time_unix_nano",):
                    assert key in data_point, "Invalid log format. Missing key: %s" % key

                all_logs.append(
                    {attr["key"]: attribute_to_value(attr["value"]) for attr in (data_point.get("attributes") or [])}
                )

        for expected_event in ml_events:
            assert expected_event in all_logs, "%s Not Found. Got: %s" % (expected_event, all_logs)

        return val

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Flask
from sanic import Sanic
import json
import webtest

from testing_support.asgi_testing import AsgiTest
from framework_graphql._target_schema_sync import target_schema as schema
from graphql_server.flask import GraphQLView as FlaskView
from graphql_server.sanic import GraphQLView as SanicView

# Sanic
target_application = dict()


def set_middlware(middleware, view_middleware):
    view_middleware.clear()
    if middleware:
        try:
            view_middleware.extend(middleware)
        except TypeError:
            view_middleware.append(middleware)


# Sanic
target_application = dict()

sanic_app = Sanic(name="SanicGraphQL")
sanic_middleware = []
sanic_view = SanicView.as_view(schema=schema, middleware=sanic_middleware)
routes = [
    sanic_app.add_route(sanic_view, "/graphql"),
]
sanic_app = AsgiTest(sanic_app)


def sanic_execute(query, middleware=None):
    set_middlware(middleware, sanic_middleware)
    response = sanic_app.make_request(
        "POST", "/graphql", body=json.dumps({"query": query}), headers={"Content-Type": "application/json"}
    )
    body = json.loads(response.body.decode("utf-8"))

    if not isinstance(query, str) or "error" in query:
        try:
            assert response.status != 200
        except AssertionError:
            assert body["errors"]
    else:
        assert response.status == 200
        assert "errors" not in body or not body["errors"]

    return response


target_application["Sanic"] = sanic_execute

# Flask

flask_app = Flask("FlaskGraphQL")
flask_middleware = []
flask_app.add_url_rule("/graphql", view_func=FlaskView.as_view("graphql", schema=schema, middleware=flask_middleware))
flask_app = webtest.TestApp(flask_app)


def flask_execute(query, middleware=None):
    if not isinstance(query, str) or "error" in query:
        expect_errors = True
    else:
        expect_errors = False

    set_middlware(middleware, flask_middleware)
    response = flask_app.post(
        "/graphql",
        json.dumps({"query": query}),
        headers={"Content-Type": "application/json"},
        expect_errors=expect_errors,
    )

    body = json.loads(response.body.decode("utf-8"))
    if expect_errors:
        assert body["errors"]
    else:
        assert "errors" not in body or not body["errors"]

    return response

target_application["Flask"] = flask_execute



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import multiprocessing
import os
import sys
import tempfile

import pytest

try:
    # python 2.x
    reload
except NameError:
    # python 3.x
    from importlib import reload


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE = os.path.normpath(os.path.join(CURRENT_DIR, "fixtures", "collector_hostname.json"))

_parameters_list = ["config_file_key", "config_override_host", "env_key", "env_override_host", "hostname"]
_parameters = ",".join(_parameters_list)


def _load_tests():
    with open(FIXTURE, "r") as fh:
        js = fh.read()
    return json.loads(js)


def _parametrize_test(test):
    return tuple([test.get(f, None) for f in _parameters_list])


_tests_json = _load_tests()
_collector_hostname_tests = [_parametrize_test(t) for t in _tests_json]
_collector_hostname_ids = [t.get("name", None) for t in _tests_json]


def _test_collector_hostname(
    config_file_key=None, config_override_host=None, env_key=None, env_override_host=None, hostname=None, queue=None
):

    try:
        ini_contents = "[newrelic]"

        if "NEW_RELIC_HOST" in os.environ:
            del os.environ["NEW_RELIC_HOST"]
        if "NEW_RELIC_LICENSE_KEY" in os.environ:
            del os.environ["NEW_RELIC_LICENSE_KEY"]

        if env_override_host:
            os.environ["NEW_RELIC_HOST"] = env_override_host
        if env_key:
            os.environ["NEW_RELIC_LICENSE_KEY"] = env_key

        if config_file_key:
            ini_contents += "\nlicense_key = %s" % config_file_key
        if config_override_host:
            ini_contents += "\nhost = %s" % config_override_host

        import newrelic.config as config
        import newrelic.core.config as core_config

        reload(core_config)
        reload(config)

        ini_file = tempfile.NamedTemporaryFile()
        ini_file.write(ini_contents.encode("utf-8"))
        ini_file.seek(0)

        config.initialize(ini_file.name)
        settings = core_config.global_settings()
        assert settings.host == hostname

    except:
        if queue:
            queue.put(sys.exc_info()[1])
        raise

    if queue:
        queue.put("PASS")


@pytest.mark.parametrize(_parameters, _collector_hostname_tests, ids=_collector_hostname_ids)
def test_collector_hostname(config_file_key, config_override_host, env_key, env_override_host, hostname):

    # We run the actual test in a subprocess because we are editing the
    # settings we need to connect to the data collector. With the wrong
    # settings, our requires_data_collector fixture will fail on any tests that
    # run after this one.

    queue = multiprocessing.Queue()
    process = multiprocessing.Process(
        target=_test_collector_hostname,
        kwargs={
            "config_file_key": config_file_key,
            "config_override_host": config_override_host,
            "env_key": env_key,
            "env_override_host": env_override_host,
            "hostname": hostname,
            "queue": queue,
        },
    )
    process.start()
    result = queue.get(timeout=2)

    assert result == "PASS"



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.in_function import wrap_in_function
from newrelic.api.wsgi_application import WSGIApplicationWrapper
from newrelic.common.package_version_utils import get_package_version


def instrument_waitress_server(module):
    def wrap_wsgi_application_entry_point(server, application, *args, **kwargs):
        dispatcher_details = ("Waitress", get_package_version("waitress"))
        application = WSGIApplicationWrapper(application, dispatcher=dispatcher_details)
        args = [server, application] + list(args)
        return (args, kwargs)

    wrap_in_function(module, "WSGIServer.__init__", wrap_wsgi_application_entry_point)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import six
import webtest

from tastypie import VERSION

from newrelic.api.background_task import background_task
from newrelic.api.transaction import end_of_transaction

from testing_support.fixtures import override_ignore_status_codes
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics

from wsgi import application

test_application = webtest.TestApp(application)

_test_application_scoped_metrics = [
        ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
        ('Function/django.http.response:HttpResponse.close', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
]

_test_application_index_scoped_metrics = list(_test_application_scoped_metrics)
_test_application_index_scoped_metrics.append(
        ('Function/views:index', 1))


@validate_code_level_metrics("views", "index")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:index',
        scoped_metrics=_test_application_index_scoped_metrics)
def test_application_index():
    response = test_application.get('/index/')
    assert response.status_code == 200
    response.mustcontain('INDEX RESPONSE')


class TastyPieFullDebugMode(object):
    def __init__(self, tastypie_full_debug):
        from django.conf import settings
        self.settings = settings
        self.tastypie_full_debug = tastypie_full_debug

    def __enter__(self):
        self.settings.TASTYPIE_FULL_DEBUG = self.tastypie_full_debug
        return 500 if self.tastypie_full_debug else 404

    def __exit__(self, *args, **kwargs):
        self.settings.TASTYPIE_FULL_DEBUG = False


_test_api_base_scoped_metrics = [
        ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
]

if six.PY3:
    _test_api_base_scoped_metrics.append(
        ('Function/tastypie.resources:Resource.wrap_view.<locals>.wrapper', 1))
else:
    _test_api_base_scoped_metrics.append(
            ('Function/tastypie.resources:wrapper', 1))

# django < 1.12 used the RegexURLResolver class and this was updated to URLResolver in later versions
if VERSION <= (0, 14, 3) and not six.PY3:
    _test_api_base_scoped_metrics.append(('Function/django.urls.resolvers:RegexURLResolver.resolve', 1))
else:
    _test_api_base_scoped_metrics.append(('Function/django.urls.resolvers:URLResolver.resolve', 1))


_test_application_not_found_scoped_metrics = list(
        _test_api_base_scoped_metrics)


@pytest.mark.parametrize('api_version', ['v1', 'v2'])
@pytest.mark.parametrize('tastypie_full_debug', [True, False])
def test_not_found(api_version, tastypie_full_debug):

    _test_application_not_found_scoped_metrics = list(
            _test_api_base_scoped_metrics)

    if tastypie_full_debug:
        _test_application_not_found_scoped_metrics.append(
                ('Function/django.http.response:HttpResponse.close', 1))
    else:
        _test_application_not_found_scoped_metrics.append(
                (('Function/django.http.response:'
                    'HttpResponseNotFound.close'), 1))

    _errors = []

    if tastypie_full_debug:
        _errors.append('tastypie.exceptions:NotFound')

    @validate_transaction_errors(errors=_errors)
    @validate_transaction_metrics('api:SimpleResource.dispatch_detail',
            scoped_metrics=_test_application_not_found_scoped_metrics)
    def _test_not_found():
        with TastyPieFullDebugMode(tastypie_full_debug) as debug_status:
            test_application.get('/api/%s/simple/NotFound/' % api_version,
                    status=debug_status)

    _test_not_found()


_test_application_object_does_not_exist_scoped_metrics = list(
        _test_api_base_scoped_metrics)

_test_application_object_does_not_exist_scoped_metrics.append(
        ('Function/tastypie.http:HttpNotFound.close', 1))


@pytest.mark.parametrize('api_version', ['v1', 'v2'])
@pytest.mark.parametrize('tastypie_full_debug', [True, False])
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('api:SimpleResource.dispatch_detail',
        scoped_metrics=_test_application_object_does_not_exist_scoped_metrics)
def test_object_does_not_exist(api_version, tastypie_full_debug):
    with TastyPieFullDebugMode(tastypie_full_debug):
        test_application.get(
                '/api/%s/simple/ObjectDoesNotExist/' % api_version, status=404)


_test_application_raises_zerodivision = list(_test_api_base_scoped_metrics)
_test_application_raises_zerodivision_exceptions = []

if six.PY3:
    _test_application_raises_zerodivision_exceptions.append(
            'builtins:ZeroDivisionError')
else:
    _test_application_raises_zerodivision_exceptions.append(
            'exceptions:ZeroDivisionError')


@pytest.mark.parametrize('api_version', ['v1', 'v2'])
@pytest.mark.parametrize('tastypie_full_debug', [True, False])
@validate_transaction_errors(
        errors=_test_application_raises_zerodivision_exceptions)
def test_raises_zerodivision(api_version, tastypie_full_debug):

    _test_application_raises_zerodivision = list(_test_api_base_scoped_metrics)

    if tastypie_full_debug:
        _test_application_raises_zerodivision.append(
                (('Function/django.core.handlers.exception:'
                    'handle_uncaught_exception'), 1))
    else:
        _test_application_raises_zerodivision.append(
                ('Function/tastypie.http:HttpApplicationError.close', 1))

    @validate_transaction_metrics('api:SimpleResource.dispatch_detail',
            scoped_metrics=_test_application_raises_zerodivision)
    def _test_raises_zerodivision():
        with TastyPieFullDebugMode(tastypie_full_debug):
            test_application.get(
                    '/api/%s/simple/ZeroDivisionError/' % api_version,
                    status=500)

    _test_raises_zerodivision()


@pytest.mark.parametrize('api_version', ['v1', 'v2'])
@pytest.mark.parametrize('tastypie_full_debug', [True, False])
@override_ignore_status_codes(set())  # don't ignore any status codes
@validate_transaction_errors(errors=['tastypie.exceptions:NotFound'])
@validate_transaction_metrics('api:SimpleResource.dispatch_detail',
        scoped_metrics=_test_application_not_found_scoped_metrics)
def test_record_404_errors(api_version, tastypie_full_debug):

    _test_application_not_found_scoped_metrics = list(
            _test_api_base_scoped_metrics)

    if tastypie_full_debug:
        _test_application_not_found_scoped_metrics.append(
                ('Function/django.http.response:HttpResponse.close', 1))
    else:
        _test_application_not_found_scoped_metrics.append(
                (('Function/django.http.response:'
                    'HttpResponseNotFound.close'), 1))

    @validate_transaction_metrics('api:SimpleResource.dispatch_detail',
            scoped_metrics=_test_application_not_found_scoped_metrics)
    def _test_not_found():
        with TastyPieFullDebugMode(tastypie_full_debug) as debug_status:
            test_application.get('/api/%s/simple/NotFound/' % api_version,
                    status=debug_status)

    _test_not_found()


@pytest.mark.parametrize('api_version', ['v1', 'v2'])
@pytest.mark.parametrize('tastypie_full_debug', [True, False])
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('test_application:test_ended_txn_name',
        background_task=True)
@background_task()
def test_ended_txn_name(api_version, tastypie_full_debug):
    # if the transaction has ended, do not change the transaction name
    end_of_transaction()

    with TastyPieFullDebugMode(tastypie_full_debug) as debug_status:
        test_application.get('/api/%s/simple/NotFound/' % api_version,
                status=debug_status)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

from gunicorn.workers.sync import SyncWorker


class WsgiProxy(object):
    def __init__(self, asgi):
        self.asgi = asgi
        self.status_code = None
        self.headers = None
        self.iterable = []

    def __call__(self, environ, start_response):
        loop = asyncio.new_event_loop()
        instance = self.asgi({"type": "http"})
        loop.run_until_complete(instance(self.receive, self.send))
        start_response("200 OK", [])
        return [b"PONG"]

    async def send(self, message):
        pass

    async def receive(self):
        return {"type": "http.request"}


class AsgiWorker(SyncWorker):
    def handle_request(self, *args, **kwargs):
        asgi = self.wsgi
        self.wsgi = WsgiProxy(asgi)

        try:
            return super(AsgiWorker, self).handle_request(*args, **kwargs)
        finally:
            self.wsgi = asgi



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import socket

def _to_int(version_str):
    m = re.match(r'\d+', version_str)
    return int(m.group(0)) if m else 0

def version2tuple(version_str):
    """Convert version, even if it contains non-numeric chars.

    >>> version2tuple('9.4rc1.1')
    (9, 4)

    """

    parts = version_str.split('.')[:2]
    return tuple(map(_to_int, parts))

def instance_hostname(hostname):
    if hostname == 'localhost' or hostname == "127.0.0.1":
        hostname = socket.gethostname()
    return hostname

def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    port = s.getsockname()[1]
    s.close()
    return port

def conditional_decorator(condition, decorator):
    """Applies a decorator if the condition is true. Accepts 0 argument callables for the condition."""
    def _conditional_decorator(func):
        if callable(condition):
            condition_eval = condition()
        else:
            condition_eval = condition
        
        if condition_eval:
            return decorator(func)
        else:
            return func

    return _conditional_decorator



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import transient_function_wrapper, function_wrapper


def validate_sql_obfuscation(expected_sqls):

    actual_sqls = []

    @transient_function_wrapper("newrelic.core.database_node", "DatabaseNode.__new__")
    def _validate_sql_obfuscation(wrapped, instance, args, kwargs):
        node = wrapped(*args, **kwargs)
        actual_sqls.append(node.formatted)
        return node

    @function_wrapper
    def _validate(wrapped, instance, args, kwargs):
        new_wrapper = _validate_sql_obfuscation(wrapped)
        result = new_wrapper(*args, **kwargs)

        for expected_sql in expected_sqls:
            assert expected_sql in actual_sqls, (expected_sql, actual_sqls)

        return result

    return _validate



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest

from newrelic.common.utilization import PCFUtilization

from testing_support.validators.validate_internal_metrics import validate_internal_metrics


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
INITIAL_ENV = os.environ
FIXTURE = os.path.normpath(os.path.join(CURRENT_DIR, 'fixtures',
    'utilization_vendor_specific', 'pcf.json'))

_parameters_list = ['testname', 'env_vars', 'expected_vendors_hash',
        'expected_metrics']

_parameters = ','.join(_parameters_list)


def _load_tests():
    with open(FIXTURE, 'r') as fh:
        js = fh.read()
    return json.loads(js)


def _parametrize_test(test):
    return tuple([test.get(f, None) for f in _parameters_list])


_pcf_tests = [_parametrize_test(t) for t in _load_tests()]


class Environ(object):
    def __init__(self, env_dict):
        env_dict = env_dict or {}
        cleaned_env_dict = {}
        for key, val in env_dict.items():
            if val is None:
                continue
            elif not isinstance(val, str):
                cleaned_env_dict[key] = val.encode('utf-8')
            else:
                cleaned_env_dict[key] = val
        self.env_dict = cleaned_env_dict

    def __enter__(self):
        os.environ.update(self.env_dict)

    def __exit__(self, *args, **kwargs):
        os.environ.clear()
        os.environ = INITIAL_ENV


class MockResponse(object):

    def __init__(self, code, body):
        self.code = code
        self.text = body

    def raise_for_status(self):
        assert str(self.code) == '200'

    def json(self):
        return self.text


@pytest.mark.parametrize(_parameters, _pcf_tests)
def test_pcf(testname, env_vars, expected_vendors_hash, expected_metrics):

    metrics = []
    if expected_metrics:
        metrics = [(k, v.get('call_count')) for k, v in
                expected_metrics.items()]

    @validate_internal_metrics(metrics=metrics)
    def _test_pcf_data():

        env_dict = dict([(key, val['response']) for key, val in
            env_vars.items()])

        with Environ(env_dict):
            data = PCFUtilization.detect()

        if data:
            pcf_vendor_hash = {'pcf': data}
        else:
            pcf_vendor_hash = None

        assert pcf_vendor_hash == expected_vendors_hash

    _test_pcf_data()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.application import application_instance
from newrelic.api.background_task import BackgroundTask
from newrelic.api.function_trace import FunctionTrace


def test_nested_context_managers():
    app = application_instance()
    outer = BackgroundTask(app, 'outer')
    inner = BackgroundTask(app, 'inner')
    with outer:
        with inner:
            assert not inner.enabled


def test_sentinel_exited_complete_root_exception():
    """
    This test forces a transaction to exit while it still has an active trace
    this causes an exception to be raised in TraceCache complete_root(). It
    verifies that the sentinel.exited property is set to true if an exception
    is raised in complete_root(), and that the exception is caught.
    """

    txn = None
    sentinel = None
    txn = BackgroundTask(application_instance(), "Parent")
    txn.__enter__()
    sentinel = txn.root_span
    trace = FunctionTrace("trace")
    trace.__enter__()
    txn.__exit__(None, None, None)
    assert sentinel.exited
    # Make sure to exit properly so cleanup is performed
    trace.__exit__(None, None, None)
    txn.__exit__(None, None, None)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import (
    function_wrapper,
    transient_function_wrapper,
)
from newrelic.core.stats_engine import CustomMetrics
from newrelic.core.internal_metrics import InternalTraceContext


def validate_internal_metrics(metrics=None):
    metrics = metrics or []

    def no_op(wrapped, instance, args, kwargs):
        pass

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        # Apply no-op wrappers to prevent new internal trace contexts from being started, preventing capture
        wrapped = transient_function_wrapper("newrelic.core.internal_metrics", "InternalTraceContext.__enter__")(no_op)(wrapped)
        wrapped = transient_function_wrapper("newrelic.core.internal_metrics", "InternalTraceContext.__exit__")(no_op)(wrapped)

        captured_metrics = CustomMetrics()
        with InternalTraceContext(captured_metrics):
            result = wrapped(*args, **kwargs)
        captured_metrics = dict(captured_metrics.metrics())

        def _validate(name, count):
            metric = captured_metrics.get(name)

            def _metrics_table():
                return "metric=%r, metrics=%r" % (name, captured_metrics)

            def _metric_details():
                return "metric=%r, count=%r" % (name, metric.call_count)

            if count is not None and count > 0:
                assert metric is not None, _metrics_table()
                if count == "present":
                    assert metric.call_count > 0, _metric_details()
                else:
                    assert metric.call_count == count, _metric_details()

            else:
                assert metric is None, _metrics_table()

        for metric, count in metrics:
            _validate(metric, count)

        return result

    return _validate_wrapper


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

import pytest

from newrelic.core.trace_cache import TraceCache

_TEST_CONCURRENT_ITERATION_TC_SIZE = 20


class DummyTrace(object):
    pass


@pytest.fixture(scope="function")
def trace_cache():
    return TraceCache()


def test_trace_cache_methods(trace_cache):
    """Test MutableMapping methods functional for trace_cache"""
    obj = DummyTrace()  # weakref compatible object

    trace_cache[1] = obj
    assert 1 in trace_cache
    assert bool(trace_cache)
    assert list(trace_cache)

    del trace_cache[1]
    assert 1 not in trace_cache
    assert not bool(trace_cache)

    trace_cache[1] = obj
    assert trace_cache.get(1, None)
    assert trace_cache.pop(1, None)

    trace_cache[1] = obj
    assert len(trace_cache) == 1
    assert len(list(trace_cache.items())) == 1
    assert len(list(trace_cache.keys())) == 1
    assert len(list(trace_cache.values())) == 1


@pytest.fixture(scope="function")
def iterate_trace_cache(trace_cache):
    def _iterate_trace_cache(shutdown):
        while True:
            if shutdown.is_set():
                return
            for k, v in trace_cache.items():
                pass
            for v in trace_cache.values():
                pass
            for v in trace_cache.keys():
                pass

    return _iterate_trace_cache


@pytest.fixture(scope="function")
def change_weakref_dict_size(trace_cache):
    def _change_weakref_dict_size(shutdown, obj_refs):
        """
        Cause RuntimeErrors when iterating on the trace_cache by:
          - Repeatedly pop and add batches of keys to cause size changes.
          - Randomly delete and replace some object refs so the weak references are deleted,
            causing the weakref dict to delete them and forcing further size changes.
        """

        dict_size_change = _TEST_CONCURRENT_ITERATION_TC_SIZE // 2  # Remove up to half of items
        while True:
            if shutdown.is_set():
                return

            # Delete and re-add keys
            for i in range(dict_size_change):
                trace_cache._cache.pop(i, None)
            for i in range(dict_size_change):
                trace_cache._cache[i] = obj_refs[i]

            # Replace every 3rd obj ref causing the WeakValueDictionary to drop it.
            for i, _ in enumerate(obj_refs[::3]):
                obj_refs[i] = DummyTrace()

    return _change_weakref_dict_size


def test_concurrent_iteration(iterate_trace_cache, change_weakref_dict_size):
    """
    Test for exceptions related to trace_cache changing size during iteration.

    The WeakValueDictionary used internally is particularly prone to this, as iterating
    on it in any way other than indirectly through WeakValueDictionary.valuerefs()
    will cause RuntimeErrors due to the unguarded iteration on a dictionary internally.
    """
    obj_refs = [DummyTrace() for _ in range(_TEST_CONCURRENT_ITERATION_TC_SIZE)]
    shutdown = threading.Event()

    t1 = threading.Thread(target=change_weakref_dict_size, args=(shutdown, obj_refs))
    t2 = threading.Thread(target=iterate_trace_cache, args=(shutdown,))
    t1.daemon = True
    t2.daemon = True
    t1.start()
    t2.start()

    # Run for 1 second, then shutdown. Stop immediately for exceptions.
    t2.join(timeout=1)
    assert t1.is_alive(), "Thread exited with exception."
    assert t2.is_alive(), "Thread exited with exception."
    shutdown.set()

    # Ensure threads shutdown with a timeout to prevent hangs
    t1.join(timeout=1)
    t2.join(timeout=1)
    assert not t1.is_alive(), "Thread failed to exit."
    assert not t2.is_alive(), "Thread failed to exit."



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from newrelic.common.encoding_utils import json_encode
from newrelic.common.object_wrapper import transient_function_wrapper


def validate_error_event_collector_json(num_errors=1):
    """Validate the format, types and number of errors of the data we
    send to the collector for harvest.
    """

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_error_event_collector_json(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:

            samples = list(instance.error_events)
            s_info = instance.error_events.sampling_info
            agent_run_id = 666

            # emulate the payload used in data_collector.py

            payload = (agent_run_id, s_info, samples)
            collector_json = json_encode(payload)

            decoded_json = json.loads(collector_json)

            assert decoded_json[0] == agent_run_id

            sampling_info = decoded_json[1]

            harvest_config = instance.settings.event_harvest_config
            reservoir_size = harvest_config.harvest_limits.error_event_data

            assert sampling_info["reservoir_size"] == reservoir_size
            assert sampling_info["events_seen"] == num_errors

            error_events = decoded_json[2]

            assert len(error_events) == num_errors
            for event in error_events:

                # event is an array containing intrinsics, user-attributes,
                # and agent-attributes

                assert len(event) == 3
                for d in event:
                    assert isinstance(d, dict)

        return result

    return _validate_error_event_collector_json



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

from flask import Flask
from werkzeug.exceptions import HTTPException


class CustomException(Exception):
    pass


def create_app(module):
    app = Flask(__name__)
    api = module.Api(app)


    class IndexResource(module.Resource):
        def get(self):
            return {'hello': 'world'}


    class ExceptionResource(module.Resource):
        def get(self, exception, code):
            if 'HTTPException' in exception:
                e = HTTPException()
            elif 'CustomException' in exception:
                e = CustomException()
            else:
                raise AssertionError('Unexpected exception received: %s' %
                        exception)
            e.code = code
            raise e


    api.add_resource(IndexResource, '/index', endpoint="index")
    api.add_resource(ExceptionResource, '/exception/<string:exception>/<int:code>', endpoint="exception")

    return app


def get_test_application(module, propagate_exceptions=False):
    app = create_app(module)
    app.config['PROPAGATE_EXCEPTIONS'] = propagate_exceptions
    test_application = webtest.TestApp(app)
    return test_application



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from conftest import CaplogHandler

from newrelic.api.background_task import background_task
from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_log_event_count import validate_log_event_count
from testing_support.validators.validate_log_events import validate_log_events
from testing_support.fixtures import override_application_settings



@pytest.fixture(scope="function")
def filepath_logger():
    import loguru
    _logger = loguru.logger
    caplog = CaplogHandler()
    handler_id = _logger.add(caplog, level="WARNING", format="{file}:{function} - {message}")
    _logger.caplog = caplog
    yield _logger
    del caplog.records[:]
    _logger.remove(handler_id)


@override_application_settings({
    "application_logging.local_decorating.enabled": False,
})
@reset_core_stats_engine()
def test_filepath_inspection(filepath_logger):
    # Test for regression in stack inspection that caused log messages.
    # See https://github.com/newrelic/newrelic-python-agent/issues/603

    @validate_log_events([{"message": "A", "level": "ERROR"}])
    @validate_log_event_count(1)
    @background_task()
    def test():
        filepath_logger.error("A")
        assert len(filepath_logger.caplog.records) == 1
        record = filepath_logger.caplog.records[0]
        assert record == "test_stack_inspection.py:test - A", record

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" The purpose of these tests is to confirm that using a non-standard
connection pool that does not have a `connection_kwargs` attribute
will not result in an error.
"""

import pytest
import redis
from testing_support.db_settings import redis_settings
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

DB_SETTINGS = redis_settings()[0]
REDIS_PY_VERSION = get_package_version_tuple("redis")


class FakeConnectionPool(object):
    """Connection Pool without connection_kwargs attribute."""

    def __init__(self, connection):
        self.connection = connection

    def get_connection(self, name, *keys, **options):
        return self.connection

    def release(self, connection):
        self.connection.disconnect()

    def disconnect(self):
        self.connection.disconnect()


# Settings

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}

# Metrics

# We don't record instance metrics when using redis blaster,
# so we just check for base metrics.
datastore_all_metric_count = 5 if REDIS_PY_VERSION >= (5, 0) else 3

_base_scoped_metrics = [
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/client_list", 1),
]
# client_setinfo was introduced in v5.0.0 and assigns info displayed in client_list output
if REDIS_PY_VERSION >= (5, 0):
    _base_scoped_metrics.append(
        ("Datastore/operation/Redis/client_setinfo", 2),
    )

_base_rollup_metrics = [
    ("Datastore/all", datastore_all_metric_count),
    ("Datastore/allOther", datastore_all_metric_count),
    ("Datastore/Redis/all", datastore_all_metric_count),
    ("Datastore/Redis/allOther", datastore_all_metric_count),
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/client_list", 1),
]
if REDIS_PY_VERSION >= (5, 0):
    _base_rollup_metrics.append(
        ("Datastore/operation/Redis/client_setinfo", 2),
    )

_host = instance_hostname(DB_SETTINGS["host"])
_port = DB_SETTINGS["port"]

_instance_metric_name = "Datastore/instance/Redis/%s/%s" % (_host, _port)

instance_metric_count = 5 if REDIS_PY_VERSION >= (5, 0) else 3

_enable_rollup_metrics = _base_rollup_metrics.append((_instance_metric_name, instance_metric_count))

_disable_rollup_metrics = _base_rollup_metrics.append((_instance_metric_name, None))

# Operations


def exercise_redis(client):
    client.set("key", "value")
    client.get("key")
    client.execute_command("CLIENT", "LIST", parse="LIST")


# Tests


@pytest.mark.skipif(REDIS_PY_VERSION < (2, 7), reason="Client list command introduced in 2.7")
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_custom_conn_pool:test_fake_conn_pool_enable_instance",
    scoped_metrics=_base_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_fake_conn_pool_enable_instance():
    client = redis.StrictRedis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)

    # Get a real connection

    conn = client.connection_pool.get_connection("GET")

    # Replace the original connection pool with one that doesn't
    # have the `connection_kwargs` attribute.

    fake_pool = FakeConnectionPool(conn)
    client.connection_pool = fake_pool
    assert not hasattr(client.connection_pool, "connection_kwargs")

    exercise_redis(client)


@pytest.mark.skipif(REDIS_PY_VERSION < (2, 7), reason="Client list command introduced in 2.7")
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_custom_conn_pool:test_fake_conn_pool_disable_instance",
    scoped_metrics=_base_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_fake_conn_pool_disable_instance():
    client = redis.StrictRedis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)

    # Get a real connection

    conn = client.connection_pool.get_connection("GET")

    # Replace the original connection pool with one that doesn't
    # have the `connection_kwargs` attribute.

    fake_pool = FakeConnectionPool(conn)
    client.connection_pool = fake_pool
    assert not hasattr(client.connection_pool, "connection_kwargs")

    exercise_redis(client)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.datastore_trace import wrap_datastore_trace
from newrelic.api.function_trace import wrap_function_trace

_pymongo_client_methods = (
    "save",
    "insert",
    "update",
    "drop",
    "remove",
    "find_one",
    "find",
    "count",
    "create_index",
    "ensure_index",
    "drop_indexes",
    "drop_index",
    "reindex",
    "index_information",
    "options",
    "group",
    "rename",
    "distinct",
    "map_reduce",
    "inline_map_reduce",
    "find_and_modify",
    "initialize_unordered_bulk_op",
    "initialize_ordered_bulk_op",
    "bulk_write",
    "insert_one",
    "insert_many",
    "replace_one",
    "update_one",
    "update_many",
    "delete_one",
    "delete_many",
    "find_raw_batches",
    "parallel_scan",
    "create_indexes",
    "list_indexes",
    "aggregate",
    "aggregate_raw_batches",
    "find_one_and_delete",
    "find_one_and_replace",
    "find_one_and_update",
)


def instrument_pymongo_connection(module):
    # Must name function explicitly as pymongo overrides the
    # __getattr__() method in a way that breaks introspection.

    rollup = ("Datastore/all", "Datastore/MongoDB/all")

    wrap_function_trace(
        module, "Connection.__init__", name="%s:Connection.__init__" % module.__name__, terminal=True, rollup=rollup
    )


def instrument_pymongo_mongo_client(module):
    # Must name function explicitly as pymongo overrides the
    # __getattr__() method in a way that breaks introspection.

    rollup = ("Datastore/all", "Datastore/MongoDB/all")

    wrap_function_trace(
        module, "MongoClient.__init__", name="%s:MongoClient.__init__" % module.__name__, terminal=True, rollup=rollup
    )


def instrument_pymongo_collection(module):
    def _collection_name(collection, *args, **kwargs):
        return collection.name

    for name in _pymongo_client_methods:
        if hasattr(module.Collection, name):
            wrap_datastore_trace(
                module, "Collection.%s" % name, product="MongoDB", target=_collection_name, operation=name
            )



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest

from newrelic.common.system_info import (logical_processor_count,
        physical_processor_count, total_physical_memory, physical_memory_used,
        _linux_physical_processor_count, _linux_total_physical_memory)


def test_logical_processor_count():
    assert logical_processor_count() >= 1


def test_physical_processor_count():
    processors_count, cores_count = physical_processor_count()
    assert processors_count is None or processors_count >= 0
    assert cores_count is None or cores_count >= 0


def test_total_physical_memory():
    assert total_physical_memory() >= 0


def test_physical_memory_used():
    assert physical_memory_used() >= 0

    if total_physical_memory() > 0:
        assert physical_memory_used() <= total_physical_memory()


@pytest.mark.parametrize('filename,expected', [
    ('1pack_1core_1logical.txt', (1, 1)),
    ('1pack_1core_2logical.txt', (1, 1)),
    ('1pack_2core_2logical.txt', (1, 2)),
    ('1pack_4core_4logical.txt', (1, 4)),
    ('2pack_12core_24logical.txt', (2, 12)),
    ('2pack_20core_40logical.txt', (2, 20)),
    ('2pack_2core_2logical.txt', (2, 2)),
    ('2pack_2core_4logical.txt', (2, 2)),
    ('2pack_4core_4logical.txt', (2, 4)),
    ('4pack_4core_4logical.txt', (4, 4)),
    ('8pack_8core_8logical.txt', (8, 8)),
    ('Xpack_Xcore_2logical.txt', (None, None)),
    ('malformed_file.txt', (None, None)),
    ('non-existant-file.txt', (None, None))])
def test_linux_physical_processor_count(filename, expected):
    here = os.path.dirname(__file__)
    path = os.path.join(here, 'fixtures', 'proc_cpuinfo', filename)

    result = _linux_physical_processor_count(path)
    assert result == expected


@pytest.mark.parametrize('filename,expected', [
    ('meminfo_4096MB.txt', 4194304 / 1024.0),
    ('malformed-file', None),
    ('non-existant-file.txt', None)])
def test_linux_total_physical_memory(filename, expected):
    here = os.path.dirname(__file__)
    path = os.path.join(here, 'fixtures', 'proc_meminfo', filename)

    value = _linux_total_physical_memory(path)
    assert value == expected



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.profile_trace import ProfileTraceWrapper, profile_trace


def test_profile_trace_wrapper():
    def _test():
        def nested_fn():
            pass

        nested_fn()

    wrapped_test = ProfileTraceWrapper(_test)
    wrapped_test()


@validate_transaction_metrics("test_profile_trace:test_profile_trace_empty_args", background_task=True)
@background_task()
def test_profile_trace_empty_args():
    @profile_trace()
    def _test():
        pass

    _test()


_test_profile_trace_defined_args_scoped_metrics = [("Custom/TestTrace", 1)]


@validate_transaction_metrics(
    "test_profile_trace:test_profile_trace_defined_args",
    scoped_metrics=_test_profile_trace_defined_args_scoped_metrics,
    background_task=True,
)
@background_task()
def test_profile_trace_defined_args():
    @profile_trace(name="TestTrace", group="Custom", label="Label", params={"key": "value"}, depth=7)
    def _test():
        pass

    _test()


_test_profile_trace_callable_args_scoped_metrics = [("Function/TestProfileTrace", 1)]


@validate_transaction_metrics(
    "test_profile_trace:test_profile_trace_callable_args",
    scoped_metrics=_test_profile_trace_callable_args_scoped_metrics,
    background_task=True,
)
@background_task()
def test_profile_trace_callable_args():
    def name_callable():
        return "TestProfileTrace"

    def group_callable():
        return "Function"

    def label_callable():
        return "HSM"

    def params_callable():
        return {"account_id": "12345"}

    @profile_trace(name=name_callable, group=group_callable, label=label_callable, params=params_callable, depth=0)
    def _test():
        pass

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import redis
from testing_support.db_settings import redis_cluster_settings

DB_CLUSTER_SETTINGS = redis_cluster_settings()[0]

# Set socket_timeout to 5s for fast fail, otherwise the default is to wait forever.
client = redis.RedisCluster(host=DB_CLUSTER_SETTINGS["host"], port=DB_CLUSTER_SETTINGS["port"], socket_timeout=5)

IGNORED_METHODS = {
    "MODULE_CALLBACKS",
    "MODULE_VERSION",
    "NAME",
    "add_edge",
    "add_node",
    "append_bucket_size",
    "append_capacity",
    "append_error",
    "append_expansion",
    "append_items_and_increments",
    "append_items",
    "append_max_iterations",
    "append_no_create",
    "append_no_scale",
    "append_values_and_weights",
    "append_weights",
    "batch_indexer",
    "BatchIndexer",
    "bulk",
    "call_procedure",
    "client_tracking_off",
    "client_tracking_on",
    "client",
    "close",
    "commandmixin",
    "connection_pool",
    "connection",
    "debug_segfault",
    "edges",
    "execute_command",
    "flush",
    "from_url",
    "get_connection_kwargs",
    "get_encoder",
    "get_label",
    "get_params_args",
    "get_property",
    "get_relation",
    "get_retry",
    "hscan_iter",
    "index_name",
    "labels",
    "list_keys",
    "load_document",
    "load_external_module",
    "lock",
    "name",
    "nodes",
    "parse_response",
    "pipeline",
    "property_keys",
    "register_script",
    "relationship_types",
    "response_callbacks",
    "RESPONSE_CALLBACKS",
    "sentinel",
    "set_file",
    "set_path",
    "set_response_callback",
    "set_retry",
    "transaction",
    "version",
    "ALL_NODES",
    "CLUSTER_COMMANDS_RESPONSE_CALLBACKS",
    "COMMAND_FLAGS",
    "DEFAULT_NODE",
    "ERRORS_ALLOW_RETRY",
    "NODE_FLAGS",
    "PRIMARIES",
    "RANDOM",
    "REPLICAS",
    "RESULT_CALLBACKS",
    "RedisClusterRequestTTL",
    "SEARCH_COMMANDS",
    "client_no_touch",
    "cluster_addslotsrange",
    "cluster_bumpepoch",
    "cluster_delslotsrange",
    "cluster_error_retry_attempts",
    "cluster_flushslots",
    "cluster_links",
    "cluster_myid",
    "cluster_myshardid",
    "cluster_replicas",
    "cluster_response_callbacks",
    "cluster_setslot_stable",
    "cluster_shards",
    "command_flags",
    "commands_parser",
    "determine_slot",
    "disconnect_connection_pools",
    "encoder",
    "get_default_node",
    "get_node",
    "get_node_from_key",
    "get_nodes",
    "get_primaries",
    "get_random_node",
    "get_redis_connection",
    "get_replicas",
    "keyslot",
    "mget_nonatomic",
    "monitor",
    "mset_nonatomic",
    "node_flags",
    "nodes_manager",
    "on_connect",
    "pubsub",
    "read_from_replicas",
    "reinitialize_counter",
    "reinitialize_steps",
    "replace_default_node",
    "result_callbacks",
    "set_default_node",
    "user_on_connect_func",
}

REDIS_MODULES = {
    "bf",
    "cf",
    "cms",
    "ft",
    "graph",
    "json",
    "tdigest",
    "topk",
    "ts",
}

IGNORED_METHODS |= REDIS_MODULES


def test_uninstrumented_methods():
    methods = {m for m in dir(client) if not m[0] == "_"}
    is_wrapped = lambda m: hasattr(getattr(client, m), "__wrapped__")
    uninstrumented = {m for m in methods - IGNORED_METHODS if not is_wrapped(m)}

    for module in REDIS_MODULES:
        if hasattr(client, module):
            module_client = getattr(client, module)()
            module_methods = {m for m in dir(module_client) if not m[0] == "_"}
            is_wrapped = lambda m: hasattr(getattr(module_client, m), "__wrapped__")
            uninstrumented |= {m for m in module_methods - IGNORED_METHODS if not is_wrapped(m)}

    assert not uninstrumented, "Uninstrumented methods: %s" % sorted(uninstrumented)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import transient_function_wrapper
from testing_support.fixtures import catch_background_exceptions


def validate_custom_parameters(required_params=None, forgone_params=None):
    required_params = required_params or []
    forgone_params = forgone_params or []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    @catch_background_exceptions
    def _validate_custom_parameters(wrapped, instance, args, kwargs):
        def _bind_params(transaction, *args, **kwargs):
            return transaction

        transaction = _bind_params(*args, **kwargs)

        # these are pre-destination applied attributes, so they may not
        # actually end up in a transaction/error trace, we are merely testing
        # for presence on the TransactionNode

        attrs = {}
        for attr in transaction.user_attributes:
            attrs[attr.name] = attr.value

        for name, value in required_params:
            assert name in attrs, "name=%r, params=%r" % (name, attrs)
            assert attrs[name] == value, "name=%r, value=%r, params=%r" % (name, value, attrs)

        for name, value in forgone_params:
            assert name not in attrs, "name=%r, params=%r" % (name, attrs)

        return wrapped(*args, **kwargs)

    return _validate_custom_parameters


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import aredis

from newrelic.hooks.datastore_redis import _conn_attrs_to_dict, _instance_info

REDIS_PY_VERSION = aredis.VERSION

_instance_info_tests = [
    ((), {}, ("localhost", "6379", "0")),
    ((), {"host": None}, ("localhost", "6379", "0")),
    ((), {"host": ""}, ("localhost", "6379", "0")),
    ((), {"db": None}, ("localhost", "6379", "0")),
    ((), {"db": ""}, ("localhost", "6379", "0")),
    ((), {"host": "127.0.0.1", "port": 1234, "db": 2}, ("127.0.0.1", "1234", "2")),
    (("127.0.0.1", 1234, 2), {}, ("127.0.0.1", "1234", "2")),
]


class DisabledConnection(aredis.Connection):
    @staticmethod
    def connect(*args, **kwargs):
        pass


class DisabledUnixConnection(aredis.UnixDomainSocketConnection, DisabledConnection):
    pass


@pytest.mark.parametrize("args,kwargs,expected", _instance_info_tests)
def test_strict_redis_client_instance_info(args, kwargs, expected):
    r = aredis.StrictRedis(*args, **kwargs)
    conn_kwargs = r.connection_pool.connection_kwargs
    assert _instance_info(conn_kwargs) == expected

@pytest.mark.parametrize("args,kwargs,expected", _instance_info_tests)
def test_strict_redis_connection_instance_info(args, kwargs, expected):
    r = aredis.StrictRedis(*args, **kwargs)
    r.connection_pool.connection_class = DisabledConnection
    connection = r.connection_pool.get_connection("SELECT")
    try:
        conn_kwargs = _conn_attrs_to_dict(connection)
        assert _instance_info(conn_kwargs) == expected
    finally:
        r.connection_pool.release(connection)


_instance_info_from_url_tests = [
    (("redis://localhost:1234/",), {}, ("localhost", "1234", "0")),
    (("redis://localhost:1234",), {}, ("localhost", "1234", "0")),
    (("redis://user:password@localhost:6379",), {}, ("localhost", "6379", "0")),
    (("redis://localhost:6379/2",), {}, ("localhost", "6379", "2")),
    (("redis://localhost:6379",), {"db": 2}, ("localhost", "6379", "2")),
    (("redis://@127.0.0.1:6379",), {}, ("127.0.0.1", "6379", "0")),
    (("redis://:1234/",), {}, ("localhost", "1234", "0")),
    (("redis://@:1234/",), {}, ("localhost", "1234", "0")),
    (("redis://localhost:1234/garbage",), {}, ("localhost", "1234", "0")),
]

# Behavior to default port to 6379 was added in v2.7.5
# (https://github.com/redis/redis-py/commit/cfe1041bbb8a8887531810429879bffbe705b03a)
# and removed in v4.0.0b1 (https://github.com/redis/redis-py/blame/v4.0.0b1/redis/connection.py#L997)
if (3, 5, 3) >= REDIS_PY_VERSION >= (2, 7, 5):
    _instance_info_from_url_tests.append((("redis://127.0.0.1",), {}, ("127.0.0.1", "6379", "0")))

if REDIS_PY_VERSION >= (2, 10):
    _instance_info_from_url_tests.extend(
        [
            (("rediss://localhost:6379/2/",), {}, ("localhost", "6379", "2")),          # rediss: Not a typo
            (("redis://localhost:6379",), {"host": "someotherhost"}, ("localhost", "6379", "0")),
            (("redis://localhost:6379/2",), {"db": 3}, ("localhost", "6379", "2")),
            (("redis://localhost:6379/2/?db=111",), {}, ("localhost", "6379", "111")),
            (("redis://localhost:6379?db=2",), {}, ("localhost", "6379", "2")),
            (("redis://localhost:6379/2?db=111",), {}, ("localhost", "6379", "111")),
            (("unix:///path/to/socket.sock",), {}, ("localhost", "/path/to/socket.sock", "0")),
            (("unix:///path/to/socket.sock?db=2",), {}, ("localhost", "/path/to/socket.sock", "2")),
            (("unix:///path/to/socket.sock",), {"db": 2}, ("localhost", "/path/to/socket.sock", "2")),
        ]
    )


@pytest.mark.parametrize("args,kwargs,expected", _instance_info_from_url_tests)
def test_strict_redis_client_from_url(args, kwargs, expected):
    r = aredis.StrictRedis.from_url(*args, **kwargs)
    conn_kwargs = r.connection_pool.connection_kwargs
    assert _instance_info(conn_kwargs) == expected


@pytest.mark.parametrize("args,kwargs,expected", _instance_info_from_url_tests)
def test_strict_redis_connection_from_url(args, kwargs, expected):
    r = aredis.StrictRedis.from_url(*args, **kwargs)
    if r.connection_pool.connection_class is aredis.Connection:
        r.connection_pool.connection_class = DisabledConnection
    elif r.connection_pool.connection_class is aredis.UnixDomainSocketConnection:
        r.connection_pool.connection_class = DisabledUnixConnection
    else:
        assert False, r.connection_pool.connection_class
    connection = r.connection_pool.get_connection("SELECT")
    try:
        conn_kwargs = _conn_attrs_to_dict(connection)
        assert _instance_info(conn_kwargs) == expected
    finally:
        r.connection_pool.release(connection)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import memcache

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.db_settings import memcached_settings
from testing_support.util import instance_hostname

from newrelic.api.background_task import background_task
from newrelic.common.object_wrapper import wrap_function_wrapper

DB_SETTINGS = memcached_settings()[0]
MEMCACHED_ADDR = '%s:%s' % (DB_SETTINGS['host'], DB_SETTINGS['port'])

# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
}

# Metrics

_base_scoped_metrics = [
        ('Datastore/operation/Memcached/set', 1),
        ('Datastore/operation/Memcached/get', 1),
        ('Datastore/operation/Memcached/delete', 1)]

_base_rollup_metrics = [
        ('Datastore/all', 3),
        ('Datastore/allOther', 3),
        ('Datastore/Memcached/all', 3),
        ('Datastore/Memcached/allOther', 3),
        ('Datastore/operation/Memcached/set', 1),
        ('Datastore/operation/Memcached/get', 1),
        ('Datastore/operation/Memcached/delete', 1)]

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(DB_SETTINGS['host'])
_port = DB_SETTINGS['port']

_instance_metric_name = 'Datastore/instance/Memcached/%s/%s' % (_host, _port)

_enable_rollup_metrics.append(
        (_instance_metric_name, 3)
)

_disable_rollup_metrics.append(
        (_instance_metric_name, None)
)

# Query

def _exercise_db(client):
    key = DB_SETTINGS['namespace'] + 'key'
    client.set(key, 'value')
    value = client.get(key)
    client.delete(key)
    assert value == 'value'

# Tests

@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
        'test_memcache:test_set_get_delete_enable',
        scoped_metrics=_enable_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@background_task()
def test_set_get_delete_enable():
    client = memcache.Client([MEMCACHED_ADDR])
    _exercise_db(client)

@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
        'test_memcache:test_set_get_delete_disable',
        scoped_metrics=_disable_scoped_metrics,
        rollup_metrics=_disable_rollup_metrics,
        background_task=True)
@background_task()
def test_set_get_delete_disable():
    instance_info_called = [False]

    def check_instance_info_call(wrapped, instance, args, kwargs):
        instance_info_called[0] = True

        return wrapped(*args, **kwargs)

    wrap_function_wrapper('newrelic.hooks.datastore_memcache',
            '_instance_info', check_instance_info_call)
    client = memcache.Client([MEMCACHED_ADDR])
    _exercise_db(client)
    assert not instance_info_called[0], "memcached _instance_info was called"



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements functions for creating a unique identity from a name and set of tags for use in dimensional metrics.
"""

from newrelic.core.attribute import process_user_attribute


def create_metric_identity(name, tags=None):
    if tags:
        # Convert dicts to an iterable of tuples, other iterables should already be in this form
        if isinstance(tags, dict):
            tags = tags.items()  

        # Apply attribute system sanitization.
        # process_user_attribute returns (None, None) for results that fail sanitization.
        # The filter removes these results from the iterable before creating the frozenset.
        tags = frozenset(filter(lambda args: args[0] is not None, map(lambda args: process_user_attribute(*args), tags)))

    tags = tags or None  # Set empty iterables after filtering to None 

    return (name, tags)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import redis

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.db_settings import redis_settings
from testing_support.util import instance_hostname

DB_MULTIPLE_SETTINGS = redis_settings()
REDIS_PY_VERSION = get_package_version_tuple("redis")


# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
}

# Metrics

_base_scoped_metrics = [
        ('Datastore/operation/Redis/get', 1),
        ('Datastore/operation/Redis/set', 1),
        ('Datastore/operation/Redis/client_list', 1),
]
# client_setinfo was introduced in v5.0.0 and assigns info displayed in client_list output
if REDIS_PY_VERSION >= (5, 0):
    _base_scoped_metrics.append(('Datastore/operation/Redis/client_setinfo', 2),)

datastore_all_metric_count = 5 if REDIS_PY_VERSION >= (5, 0) else 3

_base_rollup_metrics = [
        ('Datastore/all', datastore_all_metric_count),
        ('Datastore/allOther', datastore_all_metric_count),
        ('Datastore/Redis/all', datastore_all_metric_count),
        ('Datastore/Redis/allOther', datastore_all_metric_count),
        ('Datastore/operation/Redis/get', 1),
        ('Datastore/operation/Redis/set', 1),
        ('Datastore/operation/Redis/client_list', 1),
]

# client_setinfo was introduced in v5.0.0 and assigns info displayed in client_list output
if REDIS_PY_VERSION >= (5, 0):
    _base_rollup_metrics.append(('Datastore/operation/Redis/client_setinfo', 2),)


if len(DB_MULTIPLE_SETTINGS) > 1:
    redis_1 = DB_MULTIPLE_SETTINGS[0]
    redis_2 = DB_MULTIPLE_SETTINGS[1]

    host_1 = instance_hostname(redis_1['host'])
    port_1 = redis_1['port']

    host_2 = instance_hostname(redis_2['host'])
    port_2 = redis_2['port']

    instance_metric_name_1 = 'Datastore/instance/Redis/%s/%s' % (host_1, port_1)
    instance_metric_name_2 = 'Datastore/instance/Redis/%s/%s' % (host_2, port_2)

    instance_metric_name_1_count = 2 if REDIS_PY_VERSION >= (5, 0) else 2
    instance_metric_name_2_count = 3 if REDIS_PY_VERSION >= (5, 0) else 1

    _enable_rollup_metrics = _base_rollup_metrics.extend([
            (instance_metric_name_1, instance_metric_name_1_count),
            (instance_metric_name_2, instance_metric_name_2_count),
    ])

    _disable_rollup_metrics = _base_rollup_metrics.extend([
            (instance_metric_name_1, None),
            (instance_metric_name_2, None),
    ])


def exercise_redis(client_1, client_2):
    client_1.set('key', 'value')
    client_1.get('key')

    client_2.execute_command('CLIENT', 'LIST', parse='LIST')

@pytest.mark.skipif(len(DB_MULTIPLE_SETTINGS) < 2,
        reason='Test environment not configured with multiple databases.')
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics('test_multiple_dbs:test_multiple_datastores_enabled',
        scoped_metrics=_base_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@background_task()
def test_multiple_datastores_enabled():
    redis1 = DB_MULTIPLE_SETTINGS[0]
    redis2 = DB_MULTIPLE_SETTINGS[1]

    client_1 = redis.StrictRedis(host=redis1['host'], port=redis1['port'], db=0)
    client_2 = redis.StrictRedis(host=redis2['host'], port=redis2['port'], db=1)
    exercise_redis(client_1, client_2)

@pytest.mark.skipif(len(DB_MULTIPLE_SETTINGS) < 2,
        reason='Test environment not configured with multiple databases.')
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics('test_multiple_dbs:test_multiple_datastores_disabled',
        scoped_metrics=_base_scoped_metrics,
        rollup_metrics=_disable_rollup_metrics,
        background_task=True)
@background_task()
def test_multiple_datastores_disabled():
    redis1 = DB_MULTIPLE_SETTINGS[0]
    redis2 = DB_MULTIPLE_SETTINGS[1]

    client_1 = redis.StrictRedis(host=redis1['host'], port=redis1['port'], db=0)
    client_2 = redis.StrictRedis(host=redis2['host'], port=redis2['port'], db=1)
    exercise_redis(client_1, client_2)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import (
    check_error_attributes,
    core_application_stats_engine_error,
)

from newrelic.common.object_wrapper import transient_function_wrapper


def validate_error_trace_attributes_outside_transaction(
    err_name, required_params=None, forgone_params=None, exact_attrs=None
):
    required_params = required_params or {}
    forgone_params = forgone_params or {}
    exact_attrs = exact_attrs or {}

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.notice_error")
    def _validate_error_trace_attributes_outside_transaction(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise

        target_error = core_application_stats_engine_error(err_name)

        check_error_attributes(
            target_error.parameters, required_params, forgone_params, exact_attrs, is_transaction=False
        )

        return result

    return _validate_error_trace_attributes_outside_transaction



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile

import pytest

# these will be reloaded for each test
import newrelic.config
import newrelic.core.config
from newrelic.common.object_wrapper import function_wrapper
from newrelic.config import initialize
from newrelic.core.agent_protocol import AgentProtocol

# the specific methods imported here will not be changed when the modules are
# reloaded
from newrelic.core.config import (
    _environ_as_float,
    _environ_as_int,
    _remove_ignored_configs,
    finalize_application_settings,
    global_settings,
)

try:
    # python 2.x
    reload
except NameError:
    # python 3.x
    from importlib import reload

INI_FILE_WITHOUT_UTIL_CONF = b"""
[newrelic]
"""

INI_FILE_WITH_UTIL_CONF = b"""
[newrelic]

utilization.billing_hostname = file-hostname
"""

INI_FILE_WITH_BAD_UTIL_CONF = b"""
[newrelic]

utilization.billing_hostname = file-hostname
utilization.logical_processors = not-a-number
utilization.total_ram_mib = 12345
"""

ENV_WITHOUT_UTIL_CONF = {}
ENV_WITH_UTIL_CONF = {"NEW_RELIC_UTILIZATION_BILLING_HOSTNAME": "env-hostname"}
ENV_WITH_BAD_UTIL_CONF = {
    "NEW_RELIC_UTILIZATION_LOGICAL_PROCESSORS": "notanum",
    "NEW_RELIC_UTILIZATION_BILLING_HOSTNAME": "env-hostname",
    "NEW_RELIC_UTILIZATION_TOTAL_RAM_MIB": "98765",
}
ENV_WITH_HEROKU = {
    "NEW_RELIC_HEROKU_USE_DYNO_NAMES": "false",
    "NEW_RELIC_HEROKU_DYNO_NAME_PREFIXES_TO_SHORTEN": "meow wruff",
}

INITIAL_ENV = os.environ

# Tests for loading settings and testing for values precedence


class Environ(object):
    def __init__(self, env_dict):
        self.env_dict = {}
        for key in env_dict.keys():
            self.env_dict[key] = str(env_dict[key])

    def __enter__(self):
        os.environ.update(self.env_dict)

    def __exit__(self, *args, **kwargs):
        os.environ.clear()
        os.environ = INITIAL_ENV


def reset_agent_config(ini_contents, env_dict):
    @function_wrapper
    def reset(wrapped, instance, args, kwargs):
        with Environ(env_dict):
            ini_file = tempfile.NamedTemporaryFile()
            ini_file.write(ini_contents)
            ini_file.seek(0)

            # clean settings cache and reload env vars
            # Note that reload can at times work in unexpected ways. All that
            # is required here is that the globals (such as
            # newrelic.core.config._settings) be reset.
            #
            # From python docs (2.x and 3.x)
            # "When a module is reloaded, its dictionary (containing the
            # module's global variables) is retained. Redefinitions of names
            # will override the old definitions, so this is generally not a
            # problem."
            reload(newrelic.core.config)
            reload(newrelic.config)
            initialize(ini_file.name)
            returned = wrapped(*args, **kwargs)

        return returned

    return reset


@reset_agent_config(INI_FILE_WITHOUT_UTIL_CONF, ENV_WITHOUT_UTIL_CONF)
def test_otlp_host_port_default():
    settings = global_settings()
    assert settings.otlp_host == "otlp.nr-data.net"
    assert settings.otlp_port == 0


@reset_agent_config(
    INI_FILE_WITHOUT_UTIL_CONF, {"NEW_RELIC_OTLP_HOST": "custom-otlp.nr-data.net", "NEW_RELIC_OTLP_PORT": 443}
)
def test_otlp_port_override():
    settings = global_settings()
    assert settings.otlp_host == "custom-otlp.nr-data.net"
    assert settings.otlp_port == 443


@reset_agent_config(INI_FILE_WITHOUT_UTIL_CONF, ENV_WITHOUT_UTIL_CONF)
def test_heroku_default():
    settings = global_settings()
    assert settings.heroku.use_dyno_names is True
    assert settings.heroku.dyno_name_prefixes_to_shorten in (["scheduler", "run"], ["run", "scheduler"])


@reset_agent_config(INI_FILE_WITHOUT_UTIL_CONF, ENV_WITH_HEROKU)
def test_heroku_override():
    settings = global_settings()
    assert settings.heroku.use_dyno_names is False
    assert settings.heroku.dyno_name_prefixes_to_shorten in (["meow", "wruff"], ["wruff", "meow"])


@reset_agent_config(INI_FILE_WITHOUT_UTIL_CONF, ENV_WITH_UTIL_CONF)
def test_billing_hostname_from_env_vars():
    settings = global_settings()
    assert settings.utilization.billing_hostname == "env-hostname"

    (local_config,) = AgentProtocol._connect_payload("", [], [], settings)
    util_conf = local_config["utilization"].get("config")
    assert util_conf == {"hostname": "env-hostname"}


@reset_agent_config(INI_FILE_WITH_UTIL_CONF, ENV_WITH_UTIL_CONF)
def test_billing_hostname_precedence():
    # ini-file takes precedence over env vars
    settings = global_settings()
    assert settings.utilization.billing_hostname == "file-hostname"

    (local_config,) = AgentProtocol._connect_payload("", [], [], settings)
    util_conf = local_config["utilization"].get("config")
    assert util_conf == {"hostname": "file-hostname"}


@reset_agent_config(INI_FILE_WITHOUT_UTIL_CONF, ENV_WITHOUT_UTIL_CONF)
def test_billing_hostname_with_blank_ini_file_no_env():
    settings = global_settings()
    assert settings.utilization.billing_hostname is None

    # if no utilization config settings are set, the 'config' section is not in
    # the payload at all
    (local_config,) = AgentProtocol._connect_payload("", [], [], settings)
    util_conf = local_config["utilization"].get("config")
    assert util_conf is None


@reset_agent_config(INI_FILE_WITH_UTIL_CONF, ENV_WITHOUT_UTIL_CONF)
def test_billing_hostname_with_set_in_ini_not_in_env():
    settings = global_settings()
    assert settings.utilization.billing_hostname == "file-hostname"

    (local_config,) = AgentProtocol._connect_payload("", [], [], settings)
    util_conf = local_config["utilization"].get("config")
    assert util_conf == {"hostname": "file-hostname"}


@reset_agent_config(INI_FILE_WITH_BAD_UTIL_CONF, ENV_WITHOUT_UTIL_CONF)
def test_bad_value_in_ini_file():
    settings = global_settings()
    assert settings.utilization.logical_processors == 0

    (local_config,) = AgentProtocol._connect_payload("", [], [], settings)
    util_conf = local_config["utilization"].get("config")
    assert util_conf == {"hostname": "file-hostname", "total_ram_mib": 12345}


@reset_agent_config(INI_FILE_WITHOUT_UTIL_CONF, ENV_WITH_BAD_UTIL_CONF)
def test_bad_value_in_env_var():
    settings = global_settings()
    assert settings.utilization.logical_processors == 0

    (local_config,) = AgentProtocol._connect_payload("", [], [], settings)
    util_conf = local_config["utilization"].get("config")
    assert util_conf == {"hostname": "env-hostname", "total_ram_mib": 98765}


# Tests for combining with server side settings

_server_side_config_settings_util_conf = [
    {
        "foo": 123,
        "bar": 456,
        "agent_config": {"utilization.billing_hostname": "server-side-hostname"},
    },
    {
        "foo": 123,
        "bar": 456,
        "agent_config": {
            "baz": 789,
        },
    },
    {
        "foo": 123,
        "bar": 456,
    },
]


@pytest.mark.parametrize("server_settings", _server_side_config_settings_util_conf)
def test_remove_ignored_configs(server_settings):
    fixed_settings = _remove_ignored_configs(server_settings)
    agent_config = fixed_settings.get("agent_config", {})
    assert "utilization.billing_hostname" not in agent_config


@reset_agent_config(INI_FILE_WITH_UTIL_CONF, ENV_WITHOUT_UTIL_CONF)
@pytest.mark.parametrize("server_settings", _server_side_config_settings_util_conf)
def test_finalize_application_settings(server_settings):
    settings = global_settings()

    finalize_application_settings(server_side_config=server_settings, settings=settings)

    # hostname set in ini_file and not in env vars
    assert settings.utilization.billing_hostname == "file-hostname"


# Tests for _environ_as_int

_tests_environ_as_int = [
    {
        "name": "test no env var set, no default requested",
        "envvar_set": False,
        "envvar_val": None,  # None set
        "default": None,  # None requested
        "expected_value": 0,
    },
    {
        "name": "test no env var set, default requested",
        "envvar_set": False,
        "envvar_val": None,  # None set
        "default": 123,
        "expected_value": 123,
    },
    {
        "name": "test env var is not an int, no default requested",
        "envvar_set": True,
        "envvar_val": "testing",
        "default": None,  # None requested
        "expected_value": 0,
    },
    {
        "name": "test env var is not an int, default requested",
        "envvar_set": True,
        "envvar_val": "testing-more",
        "default": 1234,
        "expected_value": 1234,
    },
    {
        "name": "test env var is an int",
        "envvar_set": True,
        "envvar_val": 7239,
        "default": None,  # None requested
        "expected_value": 7239,
    },
]


_tests_environ_as_float = [
    {
        "name": "test no env var set, no default requested",
        "envvar_set": False,
        "envvar_val": None,  # None set
        "default": None,  # None requested
        "expected_value": 0.0,
    },
    {
        "name": "test no env var set, default requested",
        "envvar_set": False,
        "envvar_val": None,  # None set
        "default": 123.0,
        "expected_value": 123.0,
    },
    {
        "name": "test env var is not a float, no default requested",
        "envvar_set": True,
        "envvar_val": "testing",
        "default": None,  # None requested
        "expected_value": 0.0,
    },
    {
        "name": "test env var is not a number, default requested",
        "envvar_set": True,
        "envvar_val": "testing-more",
        "default": 1234.0,
        "expected_value": 1234.0,
    },
    {
        "name": "test env var is an int, not float",
        "envvar_set": True,
        "envvar_val": "7239",
        "default": None,  # None requested
        "expected_value": 7239.0,
    },
    {
        "name": "test env var is a float",
        "envvar_set": True,
        "envvar_val": "7239.23234",
        "default": None,  # None requested
        "expected_value": 7239.23234,
    },
]


def _test_environ(env_type, test):
    env = {"TESTING": test["envvar_val"]} if test["envvar_set"] else {}
    default = test["default"]
    with Environ(env):
        if default:
            val = env_type("TESTING", default=default)
        else:
            val = env_type("TESTING")
    assert val == test["expected_value"]


@pytest.mark.parametrize("test", _tests_environ_as_int)
def test__environ_as_int(test):
    _test_environ(_environ_as_int, test)


@pytest.mark.parametrize("test", _tests_environ_as_float)
def test__environ_as_float(test):
    _test_environ(_environ_as_float, test)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from newrelic.api.datastore_trace import DatastoreTrace, DatastoreTraceWrapper, wrap_datastore_trace
from newrelic.api.time_trace import current_trace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import wrap_function_wrapper
from newrelic.common.async_wrapper import coroutine_wrapper, async_generator_wrapper, generator_wrapper

_redis_client_sync_methods = {
    "acl_dryrun",
    "auth",
    "bgrewriteaof",
    "bitfield",
    "blmpop",
    "bzmpop",
    "client",
    "command",
    "command_docs",
    "command_getkeysandflags",
    "command_info",
    "debug_segfault",
    "expiretime",
    "failover",
    "hello",
    "latency_doctor",
    "latency_graph",
    "latency_histogram",
    "lcs",
    "lpop",
    "lpos",
    "memory_doctor",
    "memory_help",
    "monitor",
    "pexpiretime",
    "psetex",
    "psync",
    "pubsub",
    "renamenx",
    "rpop",
    "script_debug",
    "sentinel_ckquorum",
    "sentinel_failover",
    "sentinel_flushconfig",
    "sentinel_get_master_addr_by_name",
    "sentinel_master",
    "sentinel_masters",
    "sentinel_monitor",
    "sentinel_remove",
    "sentinel_reset",
    "sentinel_sentinels",
    "sentinel_set",
    "sentinel_slaves",
    "shutdown",
    "sort",
    "sort_ro",
    "spop",
    "srandmember",
    "unwatch",
    "watch",
    "zlexcount",
    "zrevrangebyscore",
}


_redis_client_async_methods = {
    "acl_cat",
    "acl_deluser",
    "acl_genpass",
    "acl_getuser",
    "acl_help",
    "acl_list",
    "acl_load",
    "acl_log_reset",
    "acl_log",
    "acl_save",
    "acl_setuser",
    "acl_users",
    "acl_whoami",
    "add_document_hash",
    "add_document",
    "add",
    "addnx",
    "aggregate",
    "aliasadd",
    "aliasdel",
    "aliasupdate",
    "alter_schema_add",
    "alter",
    "append",
    "arrappend",
    "arrindex",
    "arrinsert",
    "arrlen",
    "arrpop",
    "arrtrim",
    "bgsave",
    "bitcount",
    "bitfield_ro",
    "bitop_and",
    "bitop_not",
    "bitop_or",
    "bitop_xor",
    "bitop",
    "bitpos",
    "blmove",
    "blpop",
    "brpop",
    "brpoplpush",
    "byrank",
    "byrevrank",
    "bzpopmax",
    "bzpopmin",
    "card",
    "cdf",
    "clear",
    "client_getname",
    "client_getredir",
    "client_id",
    "client_info",
    "client_kill_filter",
    "client_kill",
    "client_list",
    "client_no_evict",
    "client_pause",
    "client_reply",
    "client_setinfo",
    "client_setname",
    "client_tracking",
    "client_trackinginfo",
    "client_unblock",
    "client_unpause",
    "cluster_add_slots",
    "cluster_addslots",
    "cluster_count_failure_report",
    "cluster_count_failure_reports",
    "cluster_count_key_in_slots",
    "cluster_countkeysinslot",
    "cluster_del_slots",
    "cluster_delslots",
    "cluster_failover",
    "cluster_forget",
    "cluster_get_keys_in_slot",
    "cluster_get_keys_in_slots",
    "cluster_info",
    "cluster_keyslot",
    "cluster_meet",
    "cluster_nodes",
    "cluster_replicate",
    "cluster_reset_all_nodes",
    "cluster_reset",
    "cluster_save_config",
    "cluster_set_config_epoch",
    "cluster_setslot",
    "cluster_slaves",
    "cluster_slots",
    "cluster",
    "command_count",
    "command_getkeys",
    "command_list",
    "command",
    "commit",
    "config_get",
    "config_resetstat",
    "config_rewrite",
    "config_set",
    "config",
    "copy",
    "count",
    "create_index",
    "create",
    "createrule",
    "dbsize",
    "debug_object",
    "debug_sleep",
    "debug",
    "decr",
    "decrby",
    "delete_document",
    "delete",
    "deleterule",
    "dict_add",
    "dict_del",
    "dict_dump",
    "drop_index",
    "dropindex",
    "dump",
    "echo",
    "eval_ro",
    "eval",
    "evalsha_ro",
    "evalsha",
    "execution_plan",
    "exists",
    "expire",
    "expireat",
    "explain_cli",
    "explain",
    "fcall_ro",
    "fcall",
    "flushall",
    "flushdb",
    "forget",
    "function_delete",
    "function_dump",
    "function_flush",
    "function_kill",
    "function_list",
    "function_load",
    "function_restore",
    "function_stats",
    "gears_refresh_cluster",
    "geoadd",
    "geodist",
    "geohash",
    "geopos",
    "georadius",
    "georadiusbymember",
    "geosearch",
    "geosearchstore",
    "get",
    "getbit",
    "getdel",
    "getex",
    "getrange",
    "getset",
    "hdel",
    "hexists",
    "hget",
    "hgetall",
    "hincrby",
    "hincrbyfloat",
    "hkeys",
    "hlen",
    "hmget",
    "hmset_dict",
    "hmset",
    "hrandfield",
    "hscan_iter",
    "hscan",
    "hset",
    "hsetnx",
    "hstrlen",
    "hvals",
    "incr",
    "incrby",
    "incrbyfloat",
    "info",
    "initbydim",
    "initbyprob",
    "insert",
    "insertnx",
    "keys",
    "lastsave",
    "latency_history",
    "latency_latest",
    "latency_reset",
    "lindex",
    "linsert",
    "list",
    "llen",
    "lmove",
    "lmpop",
    "loadchunk",
    "lolwut",
    "lpush",
    "lpushx",
    "lrange",
    "lrem",
    "lset",
    "ltrim",
    "madd",
    "max",
    "memory_malloc_stats",
    "memory_purge",
    "memory_stats",
    "memory_usage",
    "merge",
    "mexists",
    "mget",
    "migrate_keys",
    "migrate",
    "min",
    "module_list",
    "module_load",
    "module_loadex",
    "module_unload",
    "move",
    "mrange",
    "mrevrange",
    "mset",
    "msetnx",
    "numincrby",
    "object_encoding",
    "object_idletime",
    "object_refcount",
    "object",
    "objkeys",
    "objlen",
    "persist",
    "pexpire",
    "pexpireat",
    "pfadd",
    "pfcount",
    "pfmerge",
    "ping",
    "profile",
    "psubscribe",
    "pttl",
    "publish",
    "pubsub_channels",
    "pubsub_numpat",
    "pubsub_numsub",
    "pubsub_shardchannels",
    "pubsub_shardnumsub",
    "punsubscribe",
    "quantile",
    "query",
    "queryindex",
    "quit",
    "randomkey",
    "range",
    "rank",
    "readonly",
    "readwrite",
    "rename",
    "replicaof",
    "reserve",
    "reset",
    "resp",
    "restore",
    "revrange",
    "revrank",
    "role",
    "rpoplpush",
    "rpush",
    "rpushx",
    "sadd",
    "save",
    "scan_iter",
    "scan",
    "scandump",
    "scard",
    "script_exists",
    "script_flush",
    "script_kill",
    "script_load",
    "sdiff",
    "sdiffstore",
    "search",
    "select",
    "set",
    "setbit",
    "setex",
    "setnx",
    "setrange",
    "sinter",
    "sintercard",
    "sinterstore",
    "sismember",
    "slaveof",
    "slowlog_get",
    "slowlog_len",
    "slowlog_reset",
    "slowlog",
    "smembers",
    "smismember",
    "smove",
    "spellcheck",
    "spublish",
    "srem",
    "sscan_iter",
    "sscan",
    "stralgo",
    "strappend",
    "strlen",
    "subscribe",
    "substr",
    "sugadd",
    "sugdel",
    "sugget",
    "suglen",
    "sunion",
    "sunionstore",
    "swapdb",
    "sync",
    "syndump",
    "synupdate",
    "tagvals",
    "tfcall_async",
    "tfcall",
    "tfunction_delete",
    "tfunction_list",
    "tfunction_load",
    "time",
    "toggle",
    "touch",
    "trimmed_mean",
    "ttl",
    "type",
    "unlink",
    "unsubscribe",
    "wait",
    "waitaof",
    "xack",
    "xadd",
    "xautoclaim",
    "xclaim",
    "xdel",
    "xgroup_create",
    "xgroup_createconsumer",
    "xgroup_del_consumer",
    "xgroup_delconsumer",
    "xgroup_destroy",
    "xgroup_set_id",
    "xgroup_setid",
    "xinfo_consumers",
    "xinfo_groups",
    "xinfo_help",
    "xinfo_stream",
    "xlen",
    "xpending_range",
    "xpending",
    "xrange",
    "xread_group",
    "xread",
    "xreadgroup",
    "xrevrange",
    "xtrim",
    "zadd",
    "zaddoption",
    "zcard",
    "zcount",
    "zdiff",
    "zdiffstore",
    "zincrby",
    "zinter",
    "zintercard",
    "zinterstore",
    "zmpop",
    "zmscore",
    "zpopmax",
    "zpopmin",
    "zrandmember",
    "zrange",
    "zrangebylex",
    "zrangebyscore",
    "zrangestore",
    "zrank",
    "zrem",
    "zremrangebylex",
    "zremrangebyrank",
    "zremrangebyscore",
    "zrevrange",
    "zrevrangebylex",
    "zrevrank",
    "zscan_iter",
    "zscan",
    "zscore",
    "zunion",
    "zunionstore",
}

_redis_client_gen_methods = {
    "scan_iter",
    "hscan_iter",
    "sscan_iter",
    "zscan_iter",
}

_redis_client_methods = _redis_client_sync_methods.union(_redis_client_async_methods)

_redis_multipart_commands = set(["client", "cluster", "command", "config", "debug", "sentinel", "slowlog", "script"])

_redis_operation_re = re.compile(r"[-\s]+")


def _conn_attrs_to_dict(connection):
    return {
        "host": getattr(connection, "host", None),
        "port": getattr(connection, "port", None),
        "path": getattr(connection, "path", None),
        "db": getattr(connection, "db", None),
    }


def _instance_info(kwargs):
    host = kwargs.get("host") or "localhost"
    port_path_or_id = str(kwargs.get("path") or kwargs.get("port", "unknown"))
    db = str(kwargs.get("db") or 0)

    return (host, port_path_or_id, db)


def _wrap_Redis_method_wrapper_(module, instance_class_name, operation):
    name = "%s.%s" % (instance_class_name, operation)
    if operation in _redis_client_gen_methods:
        async_wrapper = generator_wrapper
    else:
        async_wrapper = None

    wrap_datastore_trace(module, name, product="Redis", target=None, operation=operation, async_wrapper=async_wrapper)


def _wrap_asyncio_Redis_method_wrapper(module, instance_class_name, operation):
    def _nr_wrapper_asyncio_Redis_method_(wrapped, instance, args, kwargs):
        from redis.asyncio.client import Pipeline

        if isinstance(instance, Pipeline):
            return wrapped(*args, **kwargs)

        # Method should be run when awaited or iterated, therefore we wrap in an async wrapper.
        return DatastoreTraceWrapper(wrapped, product="Redis", target=None, operation=operation, async_wrapper=async_wrapper)(*args, **kwargs)

    name = "%s.%s" % (instance_class_name, operation)
    if operation in _redis_client_gen_methods:
        async_wrapper = async_generator_wrapper
    else:
        async_wrapper = coroutine_wrapper

    wrap_function_wrapper(module, name, _nr_wrapper_asyncio_Redis_method_)


async def wrap_async_Connection_send_command(wrapped, instance, args, kwargs):
    transaction = current_transaction()
    if not transaction:
        return await wrapped(*args, **kwargs)

    host, port_path_or_id, db = (None, None, None)

    try:
        dt = transaction.settings.datastore_tracer
        if dt.instance_reporting.enabled or dt.database_name_reporting.enabled:
            conn_kwargs = _conn_attrs_to_dict(instance)
            host, port_path_or_id, db = _instance_info(conn_kwargs)
    except Exception:
        pass

    # Older Redis clients would when sending multi part commands pass
    # them in as separate arguments to send_command(). Need to therefore
    # detect those and grab the next argument from the set of arguments.

    operation = args[0].strip().lower()

    # If it's not a multi part command, there's no need to trace it, so
    # we can return early.

    if (
        operation.split()[0] not in _redis_multipart_commands
    ):  # Set the datastore info on the DatastoreTrace containing this function call.
        trace = current_trace()

        # Find DatastoreTrace no matter how many other traces are inbetween
        while trace is not None and not isinstance(trace, DatastoreTrace):
            trace = getattr(trace, "parent", None)

        if trace is not None:
            trace.host = host
            trace.port_path_or_id = port_path_or_id
            trace.database_name = db

        return await wrapped(*args, **kwargs)

    # Convert multi args to single arg string

    if operation in _redis_multipart_commands and len(args) > 1:
        operation = "%s %s" % (operation, args[1].strip().lower())

    operation = _redis_operation_re.sub("_", operation)

    with DatastoreTrace(
        product="Redis", target=None, operation=operation, host=host, port_path_or_id=port_path_or_id, database_name=db
    ):
        return await wrapped(*args, **kwargs)


def _nr_Connection_send_command_wrapper_(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None or not args:
        return wrapped(*args, **kwargs)

    host, port_path_or_id, db = (None, None, None)

    try:
        dt = transaction.settings.datastore_tracer
        if dt.instance_reporting.enabled or dt.database_name_reporting.enabled:
            conn_kwargs = _conn_attrs_to_dict(instance)
            host, port_path_or_id, db = _instance_info(conn_kwargs)
    except:
        pass

    # Find DatastoreTrace no matter how many other traces are inbetween
    trace = current_trace()
    while trace is not None and not isinstance(trace, DatastoreTrace):
        trace = getattr(trace, "parent", None)

    if trace is not None:
        trace.host = host
        trace.port_path_or_id = port_path_or_id
        trace.database_name = db

    # Older Redis clients would when sending multi part commands pass
    # them in as separate arguments to send_command(). Need to therefore
    # detect those and grab the next argument from the set of arguments.

    operation = args[0].strip().lower()

    # If it's not a multi part command, there's no need to trace it, so
    # we can return early.

    if operation.split()[0] not in _redis_multipart_commands:
        return wrapped(*args, **kwargs)

    # Convert multi args to single arg string

    if operation in _redis_multipart_commands and len(args) > 1:
        operation = "%s %s" % (operation, args[1].strip().lower())

    operation = _redis_operation_re.sub("_", operation)

    with DatastoreTrace(
        product="Redis",
        target=None,
        operation=operation,
        host=host,
        port_path_or_id=port_path_or_id,
        database_name=db,
        source=wrapped,
    ):
        return wrapped(*args, **kwargs)


def instrument_redis_client(module):
    if hasattr(module, "StrictRedis"):
        for name in _redis_client_methods:
            if name in vars(module.StrictRedis):
                _wrap_Redis_method_wrapper_(module, "StrictRedis", name)

    if hasattr(module, "Redis"):
        for name in _redis_client_methods:
            if name in vars(module.Redis):
                _wrap_Redis_method_wrapper_(module, "Redis", name)


def instrument_asyncio_redis_client(module):
    if hasattr(module, "Redis"):
        class_ = getattr(module, "Redis")
        for operation in _redis_client_async_methods:
            if hasattr(class_, operation):
                _wrap_asyncio_Redis_method_wrapper(module, "Redis", operation)

def instrument_redis_commands_core(module):
    _instrument_redis_commands_module(module, "CoreCommands")


def instrument_redis_commands_sentinel(module):
    _instrument_redis_commands_module(module, "SentinelCommands")


def instrument_redis_commands_json_commands(module):
    _instrument_redis_commands_module(module, "JSONCommands")


def instrument_redis_commands_search_commands(module):
    _instrument_redis_commands_module(module, "SearchCommands")


def instrument_redis_commands_timeseries_commands(module):
    _instrument_redis_commands_module(module, "TimeSeriesCommands")


def instrument_redis_commands_graph_commands(module):
    _instrument_redis_commands_module(module, "GraphCommands")


def instrument_redis_commands_bf_commands(module):
    _instrument_redis_commands_module(module, "BFCommands")
    _instrument_redis_commands_module(module, "CFCommands")
    _instrument_redis_commands_module(module, "CMSCommands")
    _instrument_redis_commands_module(module, "TDigestCommands")
    _instrument_redis_commands_module(module, "TOPKCommands")


def instrument_redis_commands_cluster(module):
    _instrument_redis_commands_module(module, "RedisClusterCommands")


def _instrument_redis_commands_module(module, class_name):
    for name in _redis_client_methods:
        if hasattr(module, class_name):
            class_instance = getattr(module, class_name)
            if hasattr(class_instance, name):
                _wrap_Redis_method_wrapper_(module, class_name, name)


def instrument_redis_connection(module):
    if hasattr(module, "Connection"):
        if hasattr(module.Connection, "send_command"):
            wrap_function_wrapper(module, "Connection.send_command", _nr_Connection_send_command_wrapper_)


def instrument_asyncio_redis_connection(module):
    if hasattr(module, "Connection"):
        if hasattr(module.Connection, "send_command"):
            wrap_function_wrapper(module, "Connection.send_command", wrap_async_Connection_send_command)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import pytest
from testing_support.fixtures import (
    function_not_called,
    override_application_settings,
    reset_core_stats_engine,
)
from testing_support.validators.validate_ml_event_count import validate_ml_event_count
from testing_support.validators.validate_ml_event_payload import (
    validate_ml_event_payload,
)
from testing_support.validators.validate_ml_events import validate_ml_events
from testing_support.validators.validate_ml_events_outside_transaction import (
    validate_ml_events_outside_transaction,
)

import newrelic.core.otlp_utils
from newrelic.api.application import application_instance as application
from newrelic.api.background_task import background_task
from newrelic.api.transaction import record_ml_event
from newrelic.core.config import global_settings
from newrelic.packages import six

try:
    # python 2.x
    reload
except NameError:
    # python 3.x
    from importlib import reload

_now = time.time()

_intrinsics = {
    "type": "LabelEvent",
    "timestamp": _now,
}


@pytest.fixture(scope="session")
def core_app(collector_agent_registration):
    app = collector_agent_registration
    return app._agent.application(app.name)


@validate_ml_event_payload(
    [{"foo": "bar", "real_agent_id": "1234567", "event.domain": "newrelic.ml_events", "event.name": "InferenceEvent"}]
)
@reset_core_stats_engine()
def test_ml_event_payload_inside_transaction(core_app):
    @background_task(name="test_ml_event_payload_inside_transaction")
    def _test():
        record_ml_event("InferenceEvent", {"foo": "bar"})

    _test()
    core_app.harvest()


@validate_ml_event_payload(
    [{"foo": "bar", "real_agent_id": "1234567", "event.domain": "newrelic.ml_events", "event.name": "InferenceEvent"}]
)
@reset_core_stats_engine()
def test_ml_event_payload_outside_transaction(core_app):
    def _test():
        app = application()
        record_ml_event("InferenceEvent", {"foo": "bar"}, application=app)

    _test()
    core_app.harvest()


@pytest.mark.parametrize(
    "params,expected",
    [
        ({"foo": "bar"}, [(_intrinsics, {"foo": "bar"})]),
        ({"foo": "bar", 123: "bad key"}, [(_intrinsics, {"foo": "bar"})]),
        ({"foo": "bar", "*" * 256: "too long"}, [(_intrinsics, {"foo": "bar"})]),
    ],
    ids=["Valid key/value", "Bad key", "Value too long"],
)
@reset_core_stats_engine()
def test_record_ml_event_inside_transaction(params, expected):
    @validate_ml_events(expected)
    @background_task()
    def _test():
        record_ml_event("LabelEvent", params)

    _test()


@pytest.mark.parametrize(
    "params,expected",
    [
        ({"foo": "bar"}, [(_intrinsics, {"foo": "bar"})]),
        ({"foo": "bar", 123: "bad key"}, [(_intrinsics, {"foo": "bar"})]),
        ({"foo": "bar", "*" * 256: "too long"}, [(_intrinsics, {"foo": "bar"})]),
    ],
    ids=["Valid key/value", "Bad key", "Value too long"],
)
@reset_core_stats_engine()
def test_record_ml_event_outside_transaction(params, expected):
    @validate_ml_events_outside_transaction(expected)
    def _test():
        app = application()
        record_ml_event("LabelEvent", params, application=app)

    _test()


@reset_core_stats_engine()
@validate_ml_event_count(count=0)
@background_task()
def test_record_ml_event_inside_transaction_bad_event_type():
    record_ml_event("!@#$%^&*()", {"foo": "bar"})


@reset_core_stats_engine()
@validate_ml_event_count(count=0)
def test_record_ml_event_outside_transaction_bad_event_type():
    app = application()
    record_ml_event("!@#$%^&*()", {"foo": "bar"}, application=app)


@reset_core_stats_engine()
@validate_ml_event_count(count=0)
@background_task()
def test_record_ml_event_inside_transaction_params_not_a_dict():
    record_ml_event("ParamsListEvent", ["not", "a", "dict"])


@reset_core_stats_engine()
@validate_ml_event_count(count=0)
def test_record_ml_event_outside_transaction_params_not_a_dict():
    app = application()
    record_ml_event("ParamsListEvent", ["not", "a", "dict"], application=app)


# Tests for ML Events configuration settings


@override_application_settings({"ml_insights_events.enabled": False})
@reset_core_stats_engine()
@validate_ml_event_count(count=0)
@background_task()
def test_ml_event_settings_check_ml_insights_disabled():
    record_ml_event("FooEvent", {"foo": "bar"})


# Test that record_ml_event() methods will short-circuit.
#
# If the ml_insights_events setting is False, verify that the
# `create_ml_event()` function is not called, in order to avoid the
# event_type and attribute processing.


@override_application_settings({"ml_insights_events.enabled": False})
@reset_core_stats_engine()
@function_not_called("newrelic.api.transaction", "create_custom_event")
@background_task()
def test_transaction_create_ml_event_not_called():
    record_ml_event("FooEvent", {"foo": "bar"})


@override_application_settings({"ml_insights_events.enabled": False})
@reset_core_stats_engine()
@function_not_called("newrelic.core.application", "create_custom_event")
@background_task()
def test_application_create_ml_event_not_called():
    app = application()
    record_ml_event("FooEvent", {"foo": "bar"}, application=app)


@pytest.fixture(scope="module", autouse=True, params=["protobuf", "json"])
def otlp_content_encoding(request):
    if six.PY2 and request.param == "protobuf":
        pytest.skip("OTLP protos are not compatible with Python 2.")

    _settings = global_settings()
    prev = _settings.debug.otlp_content_encoding
    _settings.debug.otlp_content_encoding = request.param
    reload(newrelic.core.otlp_utils)
    assert newrelic.core.otlp_utils.otlp_content_setting == request.param, "Content encoding mismatch."

    yield

    _settings.debug.otlp_content_encoding = prev



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.common.object_wrapper import (
    function_wrapper,
    transient_function_wrapper,
)
from testing_support.fixtures import check_event_attributes

def validate_transaction_event_attributes(required_params=None, forgone_params=None, exact_attrs=None, index=-1):
    required_params = required_params or {}
    forgone_params = forgone_params or {}
    exact_attrs = exact_attrs or {}

    captured_events = []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _capture_transaction_events(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            event_data = instance.transaction_events
            captured_events.append(event_data)
            return result

    @function_wrapper
    def _validate_transaction_event_attributes(wrapped, instance, args, kwargs):
        _new_wrapper = _capture_transaction_events(wrapped)
        result = _new_wrapper(*args, **kwargs)

        assert captured_events, "No events captured"
        event_data = captured_events[index]
        captured_events[:] = []

        check_event_attributes(event_data, required_params, forgone_params, exact_attrs)

        return result

    return _validate_transaction_event_attributes


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.packages import six
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.fixtures import dt_enabled
from newrelic.common.object_wrapper import function_wrapper


def validate_code_level_metrics(namespace, function, py2_namespace=None, builtin=False, count=1, index=-1):
    """Verify that code level metrics are generated for a callable."""

    if six.PY2 and py2_namespace is not None:
        namespace = py2_namespace

    if builtin:
        validator = validate_span_events(
            exact_agents={"code.function": function, "code.namespace": namespace, "code.filepath": "<builtin>"},
            unexpected_agents=["code.lineno"],
            count=count,
            index=index,
        )
    else:
        validator = validate_span_events(
            exact_agents={"code.function": function, "code.namespace": namespace},
            expected_agents=["code.lineno", "code.filepath"],
            count=count,
            index=index,
        )

    @function_wrapper
    def wrapper(wrapped, instance, args, kwargs):
        validator(dt_enabled(wrapped))(*args, **kwargs)

    return wrapper


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mysql.connector
from testing_support.db_settings import mysql_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_database_trace_inputs import (
    validate_database_trace_inputs,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

DB_SETTINGS = mysql_settings()
DB_SETTINGS = DB_SETTINGS[0]
DB_NAMESPACE = DB_SETTINGS["namespace"]
DB_PROCEDURE = "hello_" + DB_NAMESPACE

mysql_version = get_package_version_tuple("mysql.connector")

if mysql_version >= (8, 0, 30):
    _connector_metric_name = "Function/mysql.connector.pooling:connect"
else:
    _connector_metric_name = "Function/mysql.connector:connect"

_test_execute_via_cursor_scoped_metrics = [
    (_connector_metric_name, 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/select" % DB_NAMESPACE, 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/insert" % DB_NAMESPACE, 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/update" % DB_NAMESPACE, 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/delete" % DB_NAMESPACE, 1),
    ("Datastore/operation/MySQL/drop", 2),
    ("Datastore/operation/MySQL/create", 2),
    ("Datastore/statement/MySQL/%s/call" % DB_PROCEDURE, 1),
    ("Datastore/operation/MySQL/commit", 2),
    ("Datastore/operation/MySQL/rollback", 1),
]

_test_execute_via_cursor_rollup_metrics = [
    ("Datastore/all", 13),
    ("Datastore/allOther", 13),
    ("Datastore/MySQL/all", 13),
    ("Datastore/MySQL/allOther", 13),
    ("Datastore/operation/MySQL/select", 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/select" % DB_NAMESPACE, 1),
    ("Datastore/operation/MySQL/insert", 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/insert" % DB_NAMESPACE, 1),
    ("Datastore/operation/MySQL/update", 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/update" % DB_NAMESPACE, 1),
    ("Datastore/operation/MySQL/delete", 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/delete" % DB_NAMESPACE, 1),
    ("Datastore/statement/MySQL/%s/call" % DB_PROCEDURE, 1),
    ("Datastore/operation/MySQL/call", 1),
    ("Datastore/operation/MySQL/drop", 2),
    ("Datastore/operation/MySQL/create", 2),
    ("Datastore/operation/MySQL/commit", 2),
    ("Datastore/operation/MySQL/rollback", 1),
    ("Datastore/instance/MySQL/%s/%s" % (instance_hostname(DB_SETTINGS["host"]), DB_SETTINGS["port"]), 12),
]


@validate_transaction_metrics(
    "test_database:test_execute_via_cursor",
    scoped_metrics=_test_execute_via_cursor_scoped_metrics,
    rollup_metrics=_test_execute_via_cursor_rollup_metrics,
    background_task=True,
)
@validate_transaction_metrics(
    "test_database:test_execute_via_cursor",
    scoped_metrics=_test_execute_via_cursor_scoped_metrics,
    rollup_metrics=_test_execute_via_cursor_rollup_metrics,
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=dict)
@background_task()
def test_execute_via_cursor(table_name):
    assert mysql_version is not None
    connection = mysql.connector.connect(
        db=DB_SETTINGS["name"],
        user=DB_SETTINGS["user"],
        passwd=DB_SETTINGS["password"],
        host=DB_SETTINGS["host"],
        port=DB_SETTINGS["port"],
    )

    cursor = connection.cursor()

    cursor.execute("""drop table if exists `%s`""" % table_name)

    cursor.execute("""create table %s """ """(a integer, b real, c text)""" % table_name)

    cursor.executemany(
        """insert into `%s` """ % table_name + """values (%(a)s, %(b)s, %(c)s)""",
        [{"a": 1, "b": 1.0, "c": "1.0"}, {"a": 2, "b": 2.2, "c": "2.2"}, {"a": 3, "b": 3.3, "c": "3.3"}],
    )

    cursor.execute("""select * from %s""" % table_name)

    for row in cursor:
        pass

    cursor.execute(
        """update `%s` """ % table_name + """set a=%(a)s, b=%(b)s, c=%(c)s where a=%(old_a)s""",
        {"a": 4, "b": 4.0, "c": "4.0", "old_a": 1},
    )

    cursor.execute("""delete from `%s` where a=2""" % table_name)

    cursor.execute("""drop procedure if exists %s""" % DB_PROCEDURE)
    cursor.execute(
        """CREATE PROCEDURE %s()
                      BEGIN
                        SELECT 'Hello World!';
                      END"""
        % DB_PROCEDURE
    )

    cursor.callproc("%s" % DB_PROCEDURE)

    connection.commit()
    connection.rollback()
    connection.commit()


_test_connect_using_alias_scoped_metrics = [
    (_connector_metric_name, 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/select" % DB_NAMESPACE, 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/insert" % DB_NAMESPACE, 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/update" % DB_NAMESPACE, 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/delete" % DB_NAMESPACE, 1),
    ("Datastore/operation/MySQL/drop", 2),
    ("Datastore/operation/MySQL/create", 2),
    ("Datastore/statement/MySQL/%s/call" % DB_PROCEDURE, 1),
    ("Datastore/operation/MySQL/commit", 2),
    ("Datastore/operation/MySQL/rollback", 1),
]

_test_connect_using_alias_rollup_metrics = [
    ("Datastore/all", 13),
    ("Datastore/allOther", 13),
    ("Datastore/MySQL/all", 13),
    ("Datastore/MySQL/allOther", 13),
    ("Datastore/operation/MySQL/select", 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/select" % DB_NAMESPACE, 1),
    ("Datastore/operation/MySQL/insert", 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/insert" % DB_NAMESPACE, 1),
    ("Datastore/operation/MySQL/update", 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/update" % DB_NAMESPACE, 1),
    ("Datastore/operation/MySQL/delete", 1),
    ("Datastore/statement/MySQL/datastore_mysql_%s/delete" % DB_NAMESPACE, 1),
    ("Datastore/statement/MySQL/%s/call" % DB_PROCEDURE, 1),
    ("Datastore/operation/MySQL/call", 1),
    ("Datastore/operation/MySQL/drop", 2),
    ("Datastore/operation/MySQL/create", 2),
    ("Datastore/operation/MySQL/commit", 2),
    ("Datastore/operation/MySQL/rollback", 1),
    ("Datastore/instance/MySQL/%s/%s" % (instance_hostname(DB_SETTINGS["host"]), DB_SETTINGS["port"]), 12),
]


@validate_transaction_metrics(
    "test_database:test_connect_using_alias",
    scoped_metrics=_test_connect_using_alias_scoped_metrics,
    rollup_metrics=_test_connect_using_alias_rollup_metrics,
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=dict)
@background_task()
def test_connect_using_alias(table_name):
    assert mysql_version is not None
    connection = mysql.connector.connect(
        db=DB_SETTINGS["name"],
        user=DB_SETTINGS["user"],
        passwd=DB_SETTINGS["password"],
        host=DB_SETTINGS["host"],
        port=DB_SETTINGS["port"],
    )

    cursor = connection.cursor()

    cursor.execute("""drop table if exists `%s`""" % table_name)

    cursor.execute("""create table %s """ """(a integer, b real, c text)""" % table_name)

    cursor.executemany(
        """insert into `%s` """ % table_name + """values (%(a)s, %(b)s, %(c)s)""",
        [{"a": 1, "b": 1.0, "c": "1.0"}, {"a": 2, "b": 2.2, "c": "2.2"}, {"a": 3, "b": 3.3, "c": "3.3"}],
    )

    cursor.execute("""select * from %s""" % table_name)

    for row in cursor:
        pass

    cursor.execute(
        """update `%s` """ % table_name + """set a=%(a)s, b=%(b)s, c=%(c)s where a=%(old_a)s""",
        {"a": 4, "b": 4.0, "c": "4.0", "old_a": 1},
    )

    cursor.execute("""delete from `%s` where a=2""" % table_name)

    cursor.execute("""drop procedure if exists %s""" % DB_PROCEDURE)
    cursor.execute(
        """CREATE PROCEDURE %s()
                      BEGIN
                        SELECT 'Hello World!';
                      END"""
        % DB_PROCEDURE
    )

    cursor.callproc("%s" % DB_PROCEDURE)

    connection.commit()
    connection.rollback()
    connection.commit()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.common.object_wrapper import function_wrapper
from testing_support.fixtures import core_application_stats_engine

def validate_application_errors(errors=None, required_params=None, forgone_params=None):
    errors = errors or []
    required_params = required_params or []
    forgone_params = forgone_params or []

    @function_wrapper
    def _validate_application_errors(wrapped, instace, args, kwargs):

        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:

            stats = core_application_stats_engine()

            app_errors = stats.error_data()

            expected = sorted(errors)
            captured = sorted([(e.type, e.message) for e in stats.error_data()])

            assert expected == captured, "expected=%r, captured=%r, errors=%r" % (expected, captured, app_errors)

            for e in app_errors:
                for name, value in required_params:
                    assert name in e.parameters["userAttributes"], "name=%r, params=%r" % (name, e.parameters)
                    assert e.parameters["userAttributes"][name] == value, "name=%r, value=%r, params=%r" % (
                        name,
                        value,
                        e.parameters,
                    )

                for name, value in forgone_params:
                    assert name not in e.parameters["userAttributes"], "name=%r, params=%r" % (name, e.parameters)

        return result

    return _validate_application_errors



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import function_wrapper
from newrelic.api.transaction import current_transaction
from testing_support.validators.validate_distributed_tracing_header import validate_distributed_tracing_header
from testing_support.validators.validate_outbound_headers import validate_outbound_headers


@function_wrapper
def validate_messagebroker_headers(wrapped, instance, args, kwargs):
    result = wrapped(*args, **kwargs)

    transaction = current_transaction()
    settings = transaction.settings

    if settings.distributed_tracing.enabled:
        validate_distributed_tracing_header()
    else:
        validate_outbound_headers(header_id='NewRelicID',
                header_transaction='NewRelicTransaction')

    return result




# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from newrelic.common.encoding_utils import json_encode
from newrelic.common.object_wrapper import transient_function_wrapper


def validate_custom_event_collector_json(num_events=1):
    """Validate the format, types and number of custom events."""

    @transient_function_wrapper("newrelic.core.application", "Application.record_transaction")
    def _validate_custom_event_collector_json(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise

        stats = instance._stats_engine
        settings = stats.settings

        agent_run_id = 666
        sampling_info = stats.custom_events.sampling_info
        samples = list(stats.custom_events)

        # Emulate the payload used in data_collector.py

        payload = (agent_run_id, sampling_info, samples)
        collector_json = json_encode(payload)

        decoded_json = json.loads(collector_json)

        decoded_agent_run_id = decoded_json[0]
        decoded_sampling_info = decoded_json[1]
        decoded_events = decoded_json[2]

        assert decoded_agent_run_id == agent_run_id
        assert decoded_sampling_info == sampling_info

        max_setting = settings.event_harvest_config.harvest_limits.custom_event_data
        assert decoded_sampling_info["reservoir_size"] == max_setting

        assert decoded_sampling_info["events_seen"] == num_events
        assert len(decoded_events) == num_events

        for (intrinsics, attributes) in decoded_events:
            assert isinstance(intrinsics, dict)
            assert isinstance(attributes, dict)

        return result

    return _validate_custom_event_collector_json



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pytest
from testing_support.fixtures import (make_cross_agent_headers,
        override_application_settings)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_event_attributes import validate_transaction_event_attributes

ENCODING_KEY = '1234567890123456789012345678901234567890'


_custom_settings = {
        'cross_process_id': '1#1',
        'encoding_key': ENCODING_KEY,
        'trusted_account_ids': [1],
        'cross_application_tracer.enabled': True,
        'distributed_tracing.enabled': False,
        'transaction_tracer.transaction_threshold': 0.0,
}


@override_application_settings(_custom_settings)
@validate_transaction_event_attributes(
    required_params={
        'agent': (), 'user': (), 'intrinsic': (),
    },
    forgone_params={
        'agent': (), 'user': (), 'intrinsic': (),
    },
    exact_attrs={
        'agent': {
            'response.status': '200',
            'response.headers.contentType': 'text/html; charset=UTF-8',
        },
        'user': {}, 'intrinsic': {},
    },
)
@pytest.mark.parametrize('manual_flush', ['flush', 'no-flush'])
def test_response_to_inbound_cat(app, manual_flush):
    payload = (
        u'1#1', u'WebTransaction/Function/app:beep',
        0, 1.23, -1,
        'dd4a810b7cb7f937', False
    )
    headers = make_cross_agent_headers(payload, ENCODING_KEY, '1#1')

    client_cross_process_id = headers['X-NewRelic-ID']
    txn_header = headers['X-NewRelic-Transaction']

    response = app.fetch('/force-cat-response/%s/%s/%s' %
            (client_cross_process_id, txn_header, manual_flush))
    assert response.code == 200
    assert 'X-NewRelic-App-Data' in list(response.headers.keys())


@validate_transaction_event_attributes(
    required_params={'agent': (), 'user': (), 'intrinsic': ()},
    forgone_params={
        'agent': ('response.headers',),
        'user': (),
        'intrinsic': (),
    },
    exact_attrs={
        'agent': {
            'request.method': 'GET',
            'response.status': '304'},
        'user': {},
        'intrinsic': {},
    },
)
@override_application_settings(_custom_settings)
def test_cat_headers_not_inserted(app):
    payload = (
        u'1#1', u'WebTransaction/Function/app:beep',
        0, 1.23, -1,
        'dd4a810b7cb7f937', False
    )
    headers = make_cross_agent_headers(payload, ENCODING_KEY, '1#1')

    client_cross_process_id = headers['X-NewRelic-ID']
    txn_header = headers['X-NewRelic-Transaction']

    response = app.fetch('/304-cat-response/%s/%s' %
            (client_cross_process_id, txn_header))
    assert response.code == 304
    assert 'X-NewRelic-App-Data' not in list(response.headers.keys())


@override_application_settings(_custom_settings)
@validate_transaction_metrics('_target_application:SimpleHandler.get',
        rollup_metrics=[('ClientApplication/1#1/all', 1)])
@validate_transaction_event_attributes(
    required_params={'agent': [], 'user': [], 'intrinsic': []},
    forgone_params={'agent': [], 'user': [], 'intrinsic': []},
    exact_attrs={'agent': {}, 'user': {},
        'intrinsic': {'nr.referringTransactionGuid': 'b854df4feb2b1f06'}},
)
def test_inbound_cat_metrics_and_intrinsics(app):
    payload = ['b854df4feb2b1f06', False, '7e249074f277923d', '5d2957be']
    headers = make_cross_agent_headers(payload, ENCODING_KEY, '1#1')

    response = app.fetch('/simple', headers=headers)
    assert response.code == 200


@override_application_settings({
    'account_id': 1,
    'trusted_account_key': 1,
    'primary_application_id': 1,
    'distributed_tracing.enabled': True,
})
@validate_transaction_metrics(
    '_target_application:SimpleHandler.get',
    rollup_metrics=(
        ('Supportability/DistributedTrace/AcceptPayload/Success', 1),
    )
)
def test_inbound_dt(app):
    PAYLOAD = {
        "v": [0, 1],
        "d": {
            "ac": 1,
            "ap": 1,
            "id": "7d3efb1b173fecfa",
            "tx": "e8b91a159289ff74",
            "pr": 1.234567,
            "sa": True,
            "ti": 1518469636035,
            "tr": "d6b4ba0c3a712ca",
            "ty": "App"
        }
    }
    headers = {'newrelic': json.dumps(PAYLOAD)}
    response = app.fetch('/simple', headers=headers)
    assert response.code == 200
